/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.netbios;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * Provides logging methods specific to the netbios package. Log class
 * members are applied to the Log's mask(set as an arithmetic expression
 * from {@link jcifs.util.Log#setMask(int mask)}) to "flip on" logging of
 * netbios specific information.
 * <p><blockquote><pre>
 *     Log.setMask( Log.EXCEPTIONS +
 *                 jcifs.netbios.Log.SESSION_SERVICE_PACKET_DATA );
 * </pre></blockquote>
 * <p>
 * See the {@link jcifs.util.Log} parent class for details about this
 * logging style.
 *
 * @version   0.1, 00/03/13
 * @author    Michael B. Allen
 * @see       jcifs.util.Log
 * @since     jcifs-0.1
 */

public class Log extends jcifs.util.Log {

    // supress javadoc constructor summary
    Log() {}

    public static final int NONE                        = 0xFFFFFF0F;
    public static final int ALL                         = 0x000000F0;

/**
 *
 * This mask produces limited netbios name service packet information. See
 * <a href="http://www.cis.ohio-state.edu/rfc/rfc1002.txt">RFC 1002</a>
 * for a detailed description of packet contents and their meaning.
 * <p><blockquote><pre>
 *  Mar 13 09:17:04 - name service packet
 *   angus.mimosa.com/192.168.1.152:137 [62]
 *   28887  Positive Name Query Response
 *   authoritative answer,unicast,recursion desired,recursion available
 * </pre></blockquote>
 *
 * @see jcifs.util.Log#setMask(int mask)
 */

    public static final int NAME_SERVICE_PACKET_DATA    = 0x00000010;

    public static final int SESSION_SERVICE_PACKET_DATA = 0x00000020;

/**
 * This mask produces the contents of the netbios address cache.
 *
 * @see jcifs.util.Log#setMask(int mask)
 */

    public static final int ADDRESS_CACHE               = 0x00000040;

/**
 * reserved for this package
 */
    public static final int RESERVED8                   = 0x00000080;

	static void printPacketData( String desc, SessionServicePacket ssp ) {
		try {
			if(( SESSION_SERVICE_PACKET_DATA & mask ) == 0 ) {
				return;
			}
            out.println( desc, ssp.toString() );
        } catch( Exception e ) {
            Log.printStackTrace( "jcifs.netbios.Log.printPacketData()", e );
        }
	}

    static void printPacketData( String desc, NameServicePacket nsp ) {
        try {
            if(( NAME_SERVICE_PACKET_DATA & mask ) == 0 ) {
                return;
			}
			out.println( desc, nsp.toString() );
        } catch( Exception e ) {
            Log.printStackTrace( "jcifs.netbios.Log.printPacketData()", e );
        }
    }
    static void printAddressCache( String desc, Hashtable addressCache ) {
        if(( ADDRESS_CACHE & mask ) == 0 )
            return;
        StringBuffer sb = new StringBuffer();
        NbtAddress.CacheEntry ce;
        for( Enumeration e = addressCache.elements(); e.hasMoreElements(); ) {
            ce = (NbtAddress.CacheEntry)e.nextElement();
            sb.append( ' ' ).append( ce.hostName );
			sb.append( ' ' ).append( ce.address ).append( NL );
        }
        out.println( desc, sb.toString() );
    }
}

