/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 *                     "Christopher R. Hertel" <crh@nts.umn.edu>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.netbios;

class Name {

    static final char[] HEX_DIGITS = { 
    '0' , '1' , '2' , '3' , '4' , '5' ,
    '6' , '7' , '8' , '9' , 'A' , 'B' ,
    'C' , 'D' , 'E' , 'F'
    };

    static String toHexChars( int val ) {
        int len = 2;
        byte[] tmp = new byte[2];
        while( len > 0 ) {
            tmp[ len - 1 ] = (byte)HEX_DIGITS[ val & 0x000F ];
            if( val != 0 ) {
                val >>>= 4;
            }
            len--;
        }
        return new String( tmp );
    }

	static final int TYPE_OFFSET = 31;
	static final int SCOPE_OFFSET = 33;

    String name, scope;
    int type;

    Name( String name, int type, String scope ) {
		this.name = name;
		this.type = type;
		this.scope = scope;
    }

	int writeWireFormat( byte[] dst, int dstIndex ) {
        // write 0x20 in first byte
        dst[dstIndex] = 0x20;

        // write name
	    name = ( name.length() > 15 ) ?
							name.substring( 0, 15 ).toUpperCase() :
							name.toUpperCase();
        byte tmp[] = name.getBytes();
        int i;
        for( i = 0; i < tmp.length; i++ ) {
            dst[dstIndex + ( 2 * i + 1 )] = (byte)((( tmp[i] & 0xF0 ) >> 4 ) + 0x41 );
            dst[dstIndex + ( 2 * i + 2 )] = (byte)(( tmp[i] & 0x0F ) + 0x41 );
        }
        for( ; i < 15; i++ ) {
            dst[dstIndex + ( 2 * i + 1 )] = (byte)0x43;
            dst[dstIndex + ( 2 * i + 2 )] = (byte)0x41;
        }
        dst[dstIndex + TYPE_OFFSET] = (byte)((( type & 0xF0 ) >> 4 ) + 0x41 );
        dst[dstIndex + TYPE_OFFSET + 1] = (byte)(( type & 0x0F ) + 0x41 );
		return SCOPE_OFFSET + writeScopeWireFormat( dst, dstIndex + SCOPE_OFFSET );
	}

	int readWireFormat( byte[] src, int srcIndex ) {

        byte tmp[] = new byte[SCOPE_OFFSET];
        int length = 15;
        for( int i = 0; i < 15; i++ ) {
            tmp[i] = (byte)((( src[srcIndex + ( 2 * i + 1 )] & 0xFF ) - 0x41 ) << 4 );
            tmp[i] |= (byte)((( src[srcIndex + ( 2 * i + 2 )] & 0xFF ) - 0x41 ) & 0x0F );
            if( tmp[i] != (byte)' ' ) {
                length = i + 1;
            }
        }
        name = new String( tmp, 0, length );
        type = (( src[srcIndex + TYPE_OFFSET] & 0xFF ) - 0x41 ) << 4;
        type |= (( src[srcIndex + TYPE_OFFSET + 1] & 0xFF ) - 0x41 ) & 0x0F;
		return SCOPE_OFFSET + readScopeWireFormat( src, srcIndex );
    }
	int writeScopeWireFormat( byte[] dst, int dstIndex ) {
		if( scope == null || scope.length() == 0 ) {
			dst[dstIndex] = (byte)0x00;
			return 1;
		}

        // copy new scope in
        dst[dstIndex++] = (byte)'.';
        System.arraycopy( scope.getBytes(), 0, dst, dstIndex, scope.length() );
		dstIndex += scope.length();

		dst[dstIndex++] = (byte)0x00;

        // now go over scope backwards converting '.' to label length

		int i = dstIndex - 2;
		int e = i - scope.length();
		int c = 0;

		do {
            if( dst[i] == '.' ) {
                dst[i] = (byte)c;
                c = 0;
            } else {
                c++;
            }
		} while( i-- > e );
		return scope.length() + 2;
    }
	int readScopeWireFormat( byte[] src, int srcIndex ) {
		int start = srcIndex;
		int n;

		if(( n = src[srcIndex++] & 0x00 ) == 0 ) {
			scope = null;
			return 1;
		}

		StringBuffer sb = new StringBuffer( new String( src, srcIndex, n ));
		srcIndex += n;
		while(( n = src[srcIndex++] & 0x00 ) != 0 ) {
			sb.append( '.' ).append( new String( src, srcIndex, n ));
			srcIndex += n;
		}
		return srcIndex - start;
	}

    public int hashCode() {
		int result = name.toUpperCase().hashCode();
		result += 65599 * type;
		if( scope != null && scope.length() != 0 ) {
			result += scope.hashCode();
		}
        return result;
    }
    public boolean equals( Object obj ) {
        if( !( obj instanceof Name )) {
            return false;
        }
        Name n = (Name)obj;
        name = ( name.length() > 15 ) ?
						name.substring( 0, 15 ).toUpperCase() :
						name.toUpperCase();
        n.name = ( n.name.length() > 15 ) ?
						n.name.substring( 0, 15 ).toUpperCase() :
						n.name.toUpperCase();
		if(( scope == null || scope.length() == 0 ) &&
				( n.scope == null || n.scope.length() == 0 )) {
			return name.equals( n.name ) && type == n.type;
		}
		return name.equals( n.name ) && type == n.type && scope.equals( n.scope );
    }
    public String toString() {
        StringBuffer sb = new StringBuffer();

        name = ( name.length() > 15 ) ?
						name.substring( 0, 15 ).toUpperCase() :
						name.toUpperCase();

        // fix MSBROWSE name
        if( name.charAt( 0 ) == 0x01 ) {
            byte b[] = name.getBytes();
            b[0] = (byte)'.';
            b[1] = (byte)'.';
            b[14] = (byte)'.';
            sb.append( new String( b ));
        } else {
            sb.append( name );
        }

        sb.append( "<" ).append( toHexChars( type )).append( ">" );
        if( scope != null && scope.length() != 0 ) {
            sb.append( "." ).append( scope );
        }
        return sb.toString();
    }

}

