/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.netbios;

import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.io.IOException;
import java.util.Hashtable;
import jcifs.util.Config;

class NameServiceClient implements Runnable {

    static final int DEFAULT_SO_TIMEOUT = 5000;
    static final int DEFAULT_RCV_BUF_SIZE = 576;
    static final int DEFAULT_SND_BUF_SIZE = 576;
	static final int NAME_SERVICE_UDP_PORT = 137;
	static final int DEFAULT_RETRY_COUNT = 2;
	static final int DEFAULT_RETRY_TIMEOUT = 3000;

	NbtAddress localhost;
    InetAddress laddr, baddr, nameserver;
    int port, soTimeout, retryCount, retryTimeout;
    int snd_buf_size, rcv_buf_size;
    byte[] snd_buf, rcv_buf;
    DatagramSocket socket;
    DatagramPacket in, out;
	Hashtable responseTable;
    Thread thread;

	NameServiceClient() {
		this( Config.getInt( "jcifs.netbios.ns.lport", 0 ),
			Config.getInetAddress( "jcifs.netbios.ns.laddr", null ));
	}
    NameServiceClient( int port, InetAddress laddr ) {
        this.port = port;
        this.laddr = laddr;

		try {
			baddr = Config.getInetAddress( "jcifs.netbios.ns.baddr",
						InetAddress.getByName( "255.255.255.255" ));

			InetAddress inet = laddr;
			if( inet == null ) {
				inet = InetAddress.getLocalHost();
			}
			String name = Config.getProperty( "jcifs.netbios.ns.hostname", null );
			if( name == null || name.length() == 0 ) {
				byte[] addr = inet.getAddress();
				name = "JCIFS" +
						( addr[2] & 0xFF ) + "_" +
						( addr[3] & 0xFF ) + "_" +
						Name.toHexChars( (int)( Math.random() * (double)0xFF ));
			}
			localhost = NbtAddress.getByName( inet.getHostAddress(),
						0x00, Config.getProperty( "jcifs.netbios.ns.scope", null ));
			localhost.hostName.name = name;
			NbtAddress.cacheAddress( localhost.hostName, localhost, NbtAddress.FOREVER );
		} catch( UnknownHostException uhe ) {
		}

		nameserver = Config.getInetAddress( "jcifs.netbios.ns.nameserver", null );
        soTimeout = Config.getInt( "jcifs.netbios.ns.soTimeout", DEFAULT_SO_TIMEOUT );
		retryCount = Config.getInt( "jcifs.netbios.ns.retryCount", DEFAULT_RETRY_COUNT );
		retryTimeout = Config.getInt( "jcifs.netbios.ns.retryTimeout", DEFAULT_RETRY_TIMEOUT );
        rcv_buf_size = Config.getInt( "jcifs.netbios.ns.rcv_buf_size", DEFAULT_RCV_BUF_SIZE );
        snd_buf_size = Config.getInt( "jcifs.netbios.ns.snd_buf_size", DEFAULT_SND_BUF_SIZE );
        rcv_buf = new byte[rcv_buf_size];
        snd_buf = new byte[snd_buf_size];
        in  = new DatagramPacket( rcv_buf, rcv_buf_size );
        out = new DatagramPacket( snd_buf, snd_buf_size, baddr, NAME_SERVICE_UDP_PORT );
		responseTable = new Hashtable( 3 );
    }

    void open() throws IOException {
        if( socket != null ) {
            return;
        }
        socket = new DatagramSocket( port, laddr );
        socket.setSoTimeout( soTimeout );
        thread = new Thread( this );
        thread.start();
    }
    void close() {
        if( socket != null ) {
            socket.close();
        	socket = null;
        }
        thread = null;
		responseTable.clear();
    }
    public void run() {
		int nameTrnId;
		NameServicePacket response;
        try {
            while( socket != null ) {
				in.setLength( rcv_buf_size );
                socket.receive( in );
				Log.printHexDump( "datagram packet received from: " +
							in.getAddress().toString(), rcv_buf, 0, in.getLength() );
				nameTrnId = NameServicePacket.readNameTrnId( rcv_buf, 0 );
				response = (NameServicePacket)responseTable.get( new Integer( nameTrnId ));
				if( response == null ) {
					continue;
				}
				response.readWireFormat( rcv_buf, 0 );
				response.received = true;
				Log.printPacketData( "nbt name service packet receviced", response );
				synchronized( response ) {
					response.notify();
				}
            }
        } catch( IOException ioe ) {
        }
        close();
    }
    void send( NameServicePacket request,
					NameServicePacket response, int timeout ) throws IOException {
        if( socket == null ) {
            open();
        }
        out.setAddress( request.addr );
        out.setLength( request.writeWireFormat( snd_buf, 0 ));
		response.received = false;
		responseTable.put( new Integer( request.nameTrnId ), response );
        synchronized( response ) {
            socket.send( out );
			Log.printHexDump( "datagram packet sent to: " +
							out.getAddress().toString(), snd_buf, 0, out.getLength() );
			Log.printPacketData( "nbt name service packet sent", request );
			try {
				response.wait( timeout );
			} catch( InterruptedException ie ) {
			}
        }
		responseTable.remove( new Integer( request.nameTrnId ));
    }
	NbtAddress getByName( Name name ) throws UnknownHostException {
		NameQueryRequest request = new NameQueryRequest( name );
		NameQueryResponse response = new NameQueryResponse();

		if( nameserver != null ) {
			request.addr = nameserver;
			request.isBroadcast = false;
		} else {
			request.addr = baddr;
		}
		int n = retryCount;
		while( n-- > 0 ) {
			try {
				send( request, response, retryTimeout );
			} catch( IOException ioe ) {
				Log.printStackTrace( "nbt name service send:", ioe );
				throw new UnknownHostException( ioe.getMessage() );
			}
			if( response.received && response.resultCode == 0 ) {
				return response.addrEntry;
			}
		}
		throw new UnknownHostException( name.name );
	}
	NbtAddress[] getNodeStatus( NbtAddress addr ) throws UnknownHostException {
		NodeStatusRequest request = new NodeStatusRequest( addr.hostName );
		NodeStatusResponse response = new NodeStatusResponse( addr );
		request.addr = addr.getInetAddress();
		int n = retryCount;
		while( n-- > 0 ) {
			try {
				send( request, response, retryTimeout );
			} catch( IOException ioe ) {
				Log.printStackTrace( "nbt name service send:", ioe );
				throw new UnknownHostException( ioe.getMessage() );
			}
			if( response.received && response.resultCode == 0 ) {
				return response.addressArray;
			}
		}
		throw new UnknownHostException( addr.hostName.name );
	}
}
