/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.netbios;

import java.net.Socket;
import java.net.InetAddress;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;

/**
 * This class represents a netbios TCP/IP socket. Under no
 * conditions will users of jcifs need not be concerned with this class as
 * netbios services are handled internally to the smb package. It is only public so the smb package can access it.
 */

public class NbtSocket extends Socket {

	private static final int SSN_SRVC_PORT = 139;
	private static final int BUFFER_SIZE = 512;

	NbtAddress address;

	public NbtSocket() {
		super();
	}
	public NbtSocket( NbtAddress address, int port ) throws IOException {
		this( address, port, null, 0 );
	}
	public NbtSocket( NbtAddress address, int port,
					InetAddress localAddr, int localPort ) throws IOException {
		super( address.getInetAddress(),
					( port == 0 ? SSN_SRVC_PORT : port ), localAddr, localPort );
		this.address = address;
		connect();
	}

	public NbtAddress getNbtAddress() {
		return address;
	}
	InputStream getRawInputStream() throws IOException {
		return super.getInputStream();
	}
	OutputStream getRawOutputStream() throws IOException {
		return super.getOutputStream();
	}
	public InputStream getInputStream() throws IOException {
		return new SocketInputStream( super.getInputStream() );
	}
	public OutputStream getOutputStream() throws IOException {
		return new SocketOutputStream( super.getOutputStream() );
	}
	public int getPort() {
		return super.getPort();
	}
	public InetAddress getLocalAddress() {
		return super.getLocalAddress();
	}
	public int getLocalPort() {
		return super.getLocalPort();
	}
	public String toString() {
		return "NbtSocket[addr=" + address +
				",port=" + super.getPort() +
				",localport=" + super.getLocalPort() + "]";
	}
	private void connect() throws IOException {
		int n;
		byte[] buffer = new byte[BUFFER_SIZE];

		InputStream in = super.getInputStream();
		OutputStream out = super.getOutputStream();

		SessionServicePacket ssp0 = new SessionRequestPacket( address.hostName,
									NbtAddress.getLocalHost().hostName );
		n = ssp0.writeWireFormat( buffer, 0 );
		out.write( buffer, 0, n );

		switch( ssp0.readPacketType( in, buffer, 0 )) {
			case SessionServicePacket.POSITIVE_SESSION_RESPONSE:
				Log.println( Log.WARNINGS,
						"session service warning",
						" session established ok with " + address );
				return;

			case SessionServicePacket.NEGATIVE_SESSION_RESPONSE:
				NegativeSessionResponsePacket ssp1 = new NegativeSessionResponsePacket();
				ssp1.readWireFormat( in, buffer, 0 );
				throw new IOException( "failed to establish session with " +
						address + ": " + ssp1.errorString );

			case SessionServicePacket.SESSION_RETARGET_RESPONSE:
				SessionRetargetResponsePacket ssp2 = new SessionRetargetResponsePacket();
				ssp2.readWireFormat( in, buffer, 0 );
				throw new NbtSessionRetargetResponseException(
							"retarget session response to: " + ssp2.retargetAddress +
							":" + ssp2.retargetPort,
							ssp2.retargetAddress, ssp2.retargetPort );
			case -1:
				throw new IOException( "failed to establish session with " +
						address + " - connection refused" );
			default:
				throw new IOException( "failed to establish session with " +
						address + " - invalid session response" );
		}
	}
	public void close() throws IOException {
		Log.println( Log.WARNINGS, "netbios socket closed", this );
		super.close();
	}
}
