/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import java.util.Enumeration;

abstract class SmbComTransaction extends ServerMessageBlock implements Enumeration {

	static final byte TRANS2_FIND_FIRST2            = (byte)0x01;
	static final byte TRANS2_FIND_NEXT2             = (byte)0x02;
	static final byte TRANS2_QUERY_FS_INFORMATION   = (byte)0x03;
	static final byte TRANS2_QUERY_PATH_INFORMATION = (byte)0x05;

	// relative to headerStart
	static final int SETUP_OFFSET        = 61;

	static final int DISCONNECT_TID      = 0x01;
	static final int ONE_WAY_TRANSACTION = 0x02;

	int totalParameterCount;
	int totalDataCount;
	int maxParameterCount;
	int maxDataCount;
	byte maxSetupCount;
	int flags = 0x00;
	int timeout = 0;
	int parameterCount;
	int parameterOffset;
	int parameterDisplacement;
	int dataCount;
	int dataOffset;
	int dataDisplacement;
	int fid;
	int setupCount = 1;
	byte subCommand;
	String name = "";
	int pad = 0;
	int pad1 = 0;
	boolean hasMore = true;
	boolean isPrimary = true;
	int maxBufferSize;

	/* Subclasses must provide totalParameterCount, totalDataCount,
	 * maxParameterCount, maxDataCount, and maxSetupCount
	 * and other stuff yet to be determined if not SMB_COM_TRANSACTION2
	 */

	SmbComTransaction() {
		command = SMB_COM_TRANSACTION2;
	}

	public boolean hasMoreElements() {
		return hasMore;
	}
	public Object nextElement() {
		if( isPrimary ) {
			isPrimary = false;

			parameterDisplacement = 0;
			dataDisplacement = 0;

			parameterOffset = SETUP_OFFSET + ( setupCount * 2 ) + 2;
			if( command == SMB_COM_TRANSACTION && isResponse() == false ) {
				parameterOffset += name.length() + 1;
			}
			pad = parameterOffset % 4;
			pad = pad == 0 ? 0 : 4 - pad;
			parameterOffset += pad;
			
			int available = maxBufferSize - parameterOffset;
			parameterCount = Math.min( totalParameterCount, available );
			available -= parameterCount;

			dataOffset = parameterOffset + parameterCount;
			pad1 = dataOffset % 4;
			pad1 = pad1 == 0 ? 0 : 4 - pad1;
			dataOffset += pad1;
			dataCount = Math.min( totalDataCount, available );
		} else {
			parameterDisplacement += parameterCount;
			dataDisplacement += dataCount;
			throw new RuntimeException( "not implemented" );
		}
		if(( parameterDisplacement + parameterCount ) == totalParameterCount &&
					( dataDisplacement + dataCount ) == totalDataCount ) {
			hasMore = false;
		}
		return this;
	}
	int writeParameterWordsWireFormat( byte[] dst, int dstIndex ) {
		int start = dstIndex;

		writeInt2( totalParameterCount, dst, dstIndex );
		dstIndex += 2;
		writeInt2( totalDataCount, dst, dstIndex );
		dstIndex += 2;
		writeInt2( maxParameterCount, dst, dstIndex );
		dstIndex += 2;
		writeInt2( maxDataCount, dst, dstIndex );
		dstIndex += 2;
		dst[dstIndex++] = maxSetupCount;
		dst[dstIndex++] = (byte)0x00;           // Reserved1
		writeInt2( flags, dst, dstIndex );
		dstIndex += 2;
		writeInt4( timeout, dst, dstIndex );
		dstIndex += 4;
		dst[dstIndex++] = (byte)0x00;           // Reserved2
		dst[dstIndex++] = (byte)0x00;
		writeInt2( parameterCount, dst, dstIndex );
		dstIndex += 2;
		writeInt2( parameterOffset, dst, dstIndex );
		dstIndex += 2;
		writeInt2( dataCount, dst, dstIndex );
		dstIndex += 2;
		writeInt2(( dataCount == 0 ? 0 : dataOffset ), dst, dstIndex );
		dstIndex += 2;
		dst[dstIndex++] = (byte)setupCount;
		dst[dstIndex++] = (byte)0x00;           // Reserved3
		dst[dstIndex++] = subCommand;           // Setup[0]
		dst[dstIndex++] = (byte)0x00;

		return dstIndex - start;
	}
	int writeBytesWireFormat( byte[] dst, int dstIndex ) {
		int start = dstIndex;

		if( command == SMB_COM_TRANSACTION && isResponse() == false ) {
			dstIndex += writeString( name, dst, dstIndex );
		}

		while( pad-- > 0 ) {
			dst[dstIndex++] = (byte)0x00;       // Pad
		}

		writeParametersWireFormat( dst, dstIndex, parameterCount );
		dstIndex += parameterCount;

		if( dataCount > 0 ) {
			while( pad-- > 0 ) {
				dst[dstIndex++] = (byte)0x00;       // Pad1
			}
			writeDataWireFormat( dst, dstIndex, dataCount );
		}

		return dstIndex - start;
	}
	int readParameterWordsWireFormat( byte[] buffer, int bufferIndex ) {
		return 0;
	}
	int readBytesWireFormat( byte[] buffer, int bufferIndex ) {
		return 0;
	}

	abstract int writeParametersWireFormat( byte[] dst, int dstIndex, int len );
	abstract int writeDataWireFormat( byte[] dst, int dstIndex, int len );
	abstract int readParametersWireFormat( byte[] buffer, int bufferIndex );
	abstract int readDataWireFormat( byte[] buffer, int bufferIndex );

	public String toString() {
		return new String( super.toString() +
			",totalParameterCount=" + totalParameterCount +
			",totalDataCount=" + totalDataCount +
			",maxParameterCount=" + maxParameterCount +
			",maxDataCount=" + maxDataCount +
			",maxSetupCount=" + (int)maxSetupCount +
			",flags=0x" + Log.getHexString( flags, 2 ) +
			",timeout=" + timeout +
			",parameterCount=" + parameterCount +
			",parameterOffset=" + parameterOffset +
			",parameterDisplacement=" + parameterDisplacement +
			",dataCount=" + dataCount +
			",dataOffset=" + dataOffset +
			",dataDisplacement=" + dataDisplacement +
			",setupCount=" + setupCount );
	}
}
