/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import jcifs.netbios.NbtAddress;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.net.URL;
import java.io.IOException;

/**
 * 
 * This class represents a file or directory on an SMB file server. All
 * methods should operate identically to those of the {@link java.io.File}
 * class. One fundamental difference is the usage of a URL to specify the
 * location of a file or diectory. All SmbFile URLs have the following
 * syntax:
 * 
 * <pre>smb://[[domain;]username[:password]@]hostname[:port]/share[/path]</pre>
 * 
 * Examples:
 * 
 * <pre>
 * smb://users-nyc;msmith:PaSsW#Rd@storage15/docs/oct/YahooEarnings887.doc
 * smb://Administrator:Fo0BarP@ss@msmith1/c/WINDOWS/Desktop/YHOO Reports.txt
 * smb://storage15/public
 * </pre>
 *
 * If may also be necessary to specify properties for jcifs to
 * function correctly. See <a href="package-summary.html">Setting JCIFS
 * Properties</a> for details.
 * 
 * <p>A second constructor argument may be specified to augment the URL
 * for better programmatic control when processing many files under
 * a common base. This is slightly different from the corresponding
 * <code>java.io.File</code> usage as '/' at the beginning of the second
 * parameter will still use the same host. The examples below illustrate
 * the resulting URLs when this second contructor argument is used.
 *
 * <p>Examples:
 *
 *  <pre>
 *   base                         relative               result
 *
 *   smb://host/share/a/b      +  c/d               -->  smb://host/share/a/b/c/d
 *   smb://host/share/foo/bar  +  /share2/zig/zag   -->  smb://host/share2/zig/zag
 *   smb://host/share/foo/bar  +  ../zip            -->  smb://host/share/foo/zip
 *   smb://host/share/foo      +  ../../othershare  -->  smb://host/othershare
 *   smb://host/share/zig/zag  +  /othershare       -->  smb://host/othershare
 *   smb://host/share/zig/zag  +  smb://foo/bar     -->  smb://foo/bar
 *   smb://host/share/zig/zag  +  null              -->  smb://host/share/zig/zag
 *   smb://host/share/zig/zag  +  /                 -->  &lt;exception: at least share must be specified&gt;
 *  </pre>
 *
 * Instances of the <code>SmbFile</code> class are immutable; that is,
 * once created, the abstract pathname represented by an SmbFile object
 * will never change.
 */

public class SmbFile {

	// these are shifted for use in flags
	static final int O_RDONLY = 0x000000;
	static final int O_WRONLY = 0x010000;
	static final int O_RDWR   = 0x020000;

	// Open Function Encoding
	static final int O_CREAT  = 0x0010;
	static final int O_EXCL   = 0x0001; // opposite meaning
	static final int O_TRUNC  = 0x0002;
	static final int O_APPEND = 0x1000; // custom

	// file attribute encoding
	static final int ATTR_READONLY   = 0x01;
	static final int ATTR_HIDDEN     = 0x02;
	static final int ATTR_SYSTEM     = 0x04;
	static final int ATTR_VOLUME     = 0x08;
	static final int ATTR_DIRECTORY  = 0x10;
	static final int ATTR_ARCHIVE    = 0x20;

	static {
		jcifs.util.Config.init();
	}

	SmbURL url;
	String path;
	SmbTree tree;
	int fid, tid, attributes, size;
	long lastModified;
	boolean opened;

/** 
 * Constructs and SmbFile representing a file or directory on an SMB server
 * specified by the URL argument. See the description and examples of smb
 * URLs above.
 */ 

	public SmbFile( String url ) throws MalformedURLException, UnknownHostException {
		this( new SmbURL( url ));
	}

/**
 * Constructs an SmbFile representing a file or directory on an SMB server
 * specified by the URL argument and the possibly null relative path. See
 * the description and examples of using the second parameter above.
 */ 

	public SmbFile( String url, String name ) throws MalformedURLException,
																UnknownHostException {
		this( new SmbURL( url, name ));
	}

/** 
 * Constructs an SmbFile representing a file or directory on an SMB
 * server specified by the SmbFile augmented with the named path. See the
 * description and examples of using the second parameter above.
 */ 

	public SmbFile( SmbFile file, String name ) throws MalformedURLException,
																UnknownHostException {
		this( new SmbURL( file.toString(), name ));
	}
	SmbFile( SmbURL url ) throws MalformedURLException, UnknownHostException {
		this.url = url;
		NbtAddress addr = NbtAddress.getByName( url.host, 0x20, "" );
		SmbTransport trans = SmbTransport.getSmbTransport( addr, url.port );
		SmbSession ssn = trans.getSmbSession( url.username, url.password, url.domain );
		String unc = new String( "\\\\" + url.host + "\\" + url.share ).toUpperCase();
		tree = ssn.getSmbTree( unc, null );
		path = url.canonicalPath;
		if( path.length() == 0 ) {
			path = "\\";
		} else {
			path = path.replace( '/', '\\' );
		}
		opened = false;
	}

	boolean isOpen() {
		if( opened && tree.treeConnected && tid == tree.tid ) {
			return true;
		}
		return false;
	}
	void open( int flags ) throws IOException {
		if( isOpen() ) {
			return;
		}
		fid = tree.open( path, flags );
		tid = tree.tid;
		opened = true;
	}
	void close() throws IOException {
		if( isOpen() == false ) {
			return;
		}
		opened = false;

		Log.println( Log.WARNINGS, "smb close warning",
				" fid=" + fid );

		/*
		 * Close Request / Response
		 */

		tree.send( new SmbComClose( fid ), new SmbComBlankResponse() );
	}

/**
 * Returns the last component of the path.
 */

	public String getName() {
		return url.file;
	}

/**
 * Returns everything but the last component of the path.
 */

	public String getParent() {
		String result = url.toString();
		int p = result.lastIndexOf( '/' );
		if( p != -1 ) {
			result = result.substring( 0, p );
		}
		return result;
	}

/**
 * Returns the full URL.
 */

	public String getPath() {
		return url.toString();
	}

/**
 * Returns the full URL with '..' and '.' components factored out.
 */


	public String getCanonicalPath() {
		return url.getCanonicalPath();
	}

/**
 * Tests to see if the file this SmbFile represents exists on the server.
 */

	public boolean exists() {
		Info info = null;
		try {
			info = tree.queryPath( path );
		} catch( Exception e ) {
			info = null;
			Log.printStackTrace( "SmbFile exception", e );
		}
		if( info == null ) {
			return false;
		}
		attributes = info.getAttributes();
		lastModified = info.getLastWriteTime();
		lastModified += tree.session.transport.server.serverTimeZone * 1000 * 60L;
		size = info.getSize();
		return true;
	}

/**
 * Tests to see if the file this SmbFile represents is marked read-only.
 */

	public boolean canRead() {
		return canWrite();
	}

/**
 * Tests to see if the file this SmbFile represents is not marked read-only.
 */

	public boolean canWrite() {
		return exists() && ( attributes & ATTR_READONLY ) == 0;
	}

/**
 * Tests to see if the file this SmbFile represents is a directory.
 */

	public boolean isDirectory() {
		return exists() && ( attributes & ATTR_DIRECTORY ) == ATTR_DIRECTORY;
	}

/**
 * Tests to see if the file this SmbFile represents is not a directory.
 */

	public boolean isFile() {
		return !isDirectory();
	}

/**
 * Tests to see if the file this SmbFile represents is marked as being hidden.
 */

	public boolean isHidden() {
		return exists() && ( attributes & ATTR_HIDDEN ) == ATTR_HIDDEN;
	}

/**
 * Returns the time the file this SmbFile represents was lastModified.
 */

	public long lastModified() {
		exists();
		return lastModified;
	}

/**
 * Returns the names of files and directorys listed under the directory
 * this SmbFile represents.
 */

	public String[] list() {
		try {
			return tree.find( path );
		} catch( Exception e ) {
			Log.printStackTrace( "SmbFile exception", e );
		}
		return null;
	}

/**
 * Changes the name of the file this SmbFile represents to the name
 * designated by the SmbFile argument(Remember SmbFiles are immutible
 * and therefore the path associated with this SmbFile object will not
 * change).
 */

	public boolean renameTo( SmbFile dest ) {
		try {
			if( dest.exists() ) {
				// otherwise we get Access Denied
				if( dest.delete() == false ) {
					return false;
				}
			}
			return tree.rename( path, dest.path );
		} catch( Exception e ) {
			Log.printStackTrace( "SmbFile exception", e );
		}
		return false;
	}

/**
 * Deletes the file on the SMB server that this SmbFile represents.
 */

	public boolean delete() {
		try {
			return tree.delete( path, isDirectory() );
		} catch( Exception e ) {
			Log.printStackTrace( "SmbFile exception", e );
		}
		return false;
	}

/**
 * Returns the length of the file this SmbFile represents.
 */

	public long length() {
		if( exists() ) {
			// will do long a little later. not hard.
			return (long)size;
		}
		return 0L;
	}

/**
 * Creates a directory with the path specified by this SmbFile.
 */

	public boolean mkdir() {
		try {
			return tree.createDirectory( path );
		} catch( Exception e ) {
			Log.printStackTrace( "SmbFile exception", e );
		}
		return false;
	}

/**
 * Returns the URL representation of the file this SmbFile represents.
 */

	public URL toURL() throws MalformedURLException {
		return new URL( url.toString() );
	}

/**
 * Computes a hashCode for this file based on the URL string.
 */

	public int hashCode() {
		// weak
		return url.toString().hashCode();
	}

/**
 * Tests to see if two SmbFile objects are equal by comparing hashCodes.
 */

	public boolean equals( Object obj ) {
		// this might be improved?
		// are two files equal if auth is different?
		return obj instanceof SmbFile && obj.hashCode() == hashCode();
	}

/**
 * Returns the string representation of this SmbFile object.
 */

	public String toString() {
		return url.toString();
	}
}
