/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import java.io.InputStream;
import java.io.IOException;
import jcifs.util.Config;

/**
 * This InputStream can read bytes from a file on an SMB file server.
 */

public class SmbFileInputStream extends InputStream {

	private SmbFile file;
	private int fp, off, readSize, openFlags;
	private byte[] tmp = new byte[1];

	/**
	 * Creates an {@link java.io.InputStream} for reading bytes from a file
	 * on an SMB server addressed by the URL parameter. See {@link
	 * jcifs.smb.SmbFile} for a detailed description and examples of
	 * the smb URL syntax.
	 */

	public SmbFileInputStream( String url ) throws IOException {
		this( new SmbFile( url ));
	}

	/**
	 * Creates an {@link java.io.InputStream} for reading bytes from a file
	 * on an SMB server represented by the {@link jcifs.smb.SmbFile} parameter.
	 */

	public SmbFileInputStream( SmbFile file ) throws IOException {
		this.file = file;
		openFlags = SmbFile.O_RDONLY;
		file.open( openFlags );
		readSize = Math.min( file.tree.session.transport.rcv_buf_size - 70,
							file.tree.session.transport.server.maxBufferSize - 70 );
	}
	SmbFileInputStream( SmbURL url ) throws IOException {
		this( new SmbFile( url ));
	}

	/**
	 * Closes this file input stream and releases any system resources
	 * associated with the stream.
	 */

	public void close() throws IOException {
		file.close();
	}

	/**
	 * Reads a byte of data from this input stream.
	 */

	public int read() throws IOException {
		// need oplocks to cache otherwise use BufferedInputStream
		if( read( tmp, 0, 1 ) == -1 ) {
			return -1;
		}
		return tmp[0] & 0xFF;
	}

	/**
	 * Reads up to b.length bytes of data from this input stream
	 * into an array of bytes.
	 */

	public int read( byte[] b ) throws IOException {
		return read( b, 0, b.length );
	}

	/**
	 * Reads up to len bytes of data from this input stream into an
	 * array of bytes.
	 */

	public int read( byte[] b, int off, int len ) throws IOException {
		if( len <= 0 ) {
			return 0;
		}
		int start = fp;

		// ensure file is open
		if( file.isOpen() == false ) {
			file.open( openFlags );
		}

		Log.println( Log.WARNINGS, "smb read warning",
				" fid=" + file.fid + ",off=" + off + ",len=" + len );

		/*
		 * Read AndX Request / Response
		 */

		SmbComReadAndXResponse response = new SmbComReadAndXResponse( b, off );

		int r, n;
		do {
			r = len > readSize ? readSize : len;
			file.tree.send( new SmbComReadAndX( file.fid, fp, r, null ), response );
			if(( n = response.dataLength ) <= 0 ) {
				return -1;
			}
			fp += n;
			len -= n;
			response.off += n;
		} while( len > 0 && n == r );

		return fp - start;
	}
}
