/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import java.io.OutputStream;
import java.io.IOException;
import jcifs.util.Config;

/**
 * This OutputStream can write bytes to a file on an SMB file server.
 */

public class SmbFileOutputStream extends OutputStream {

	private SmbFile file;
	private boolean append;
	private int openFlags, fp, writeSize;
	private byte[] tmp = new byte[1];


	/**
	 * Creates an {@link java.io.OutputStream} for writing to a file
	 * on an SMB server addressed by the URL parameter. See {@link
	 * jcifs.smb.SmbFile} for a detailed description and examples of
	 * the smb URL syntax.
	 */

	public SmbFileOutputStream( String url ) throws IOException {
		this( url, false );
	}

	/**
	 * Creates an {@link java.io.OutputStream} for writing to a file
	 * on an SMB server represented by the SmbFile parameter.
	 */

	public SmbFileOutputStream( SmbFile file ) throws IOException {
		this( file, false );
	}

	/**
	 * Creates an {@link java.io.OutputStream} for writing to a file
	 * on an SMB server addressed by the URL parameter. See {@link
	 * jcifs.smb.SmbFile} for a detailed description and examples of
	 * the smb URL syntax. If the second argument is true, then bytes
	 * will be written to the end of the file rather than the beginning.
	 */

	public SmbFileOutputStream( String url, boolean append ) throws IOException {
		this( new SmbFile( url ), append );
	}

	public SmbFileOutputStream( SmbFile file, boolean append ) throws IOException {
		this.file = file;
		this.append = append;
		if( append ) {
			openFlags = SmbFile.O_CREAT | SmbFile.O_WRONLY | SmbFile.O_APPEND;
			fp = (int)file.length();
		} else {
			openFlags = SmbFile.O_CREAT | SmbFile.O_WRONLY | SmbFile.O_TRUNC;
			fp = 0;
		}
		file.open( openFlags );
		writeSize = Math.min( file.tree.session.transport.snd_buf_size - 70,
							file.tree.session.transport.server.maxBufferSize - 70 );
	}

	/**
	 * Closes this file output stream and releases any system
	 * resources associated with this stream.
	 */

	public void close() throws IOException {
		file.close();
	}

	/**
	 * Writes the specified byte to this file output stream.
	 */

	public void write( int b ) throws IOException {
		tmp[0] = (byte)b;
		write( tmp, 0, 1 );
	}

	/**
	 * Writes b.length bytes from the specified byte array to this
	 * file output stream.
	 */

	public void write( byte[] b ) throws IOException {
		write( b, 0, b.length );
	}

	/**
	 * Writes len bytes from the specified byte array starting at
	 * offset off to this file output stream.
	 */

	public void write( byte[] b, int off, int len ) throws IOException {
		if( len <= 0 ) {
			return;
		}

		// ensure file is open
		if( file.isOpen() == false ) {
			file.open( openFlags );
			if( append ) {
				fp = (int)file.length();
			}
		}

		Log.println( Log.WARNINGS, "smb write warning",
						" fid=" + file.fid + ",off=" + off + ",len=" + len );
		int w;
		do {
			w = len > writeSize ? writeSize : len;
			file.tree.send( new SmbComWriteAndX( file.fid, fp, len - w, b, off, w, null ),
														new SmbComWriteAndXResponse() );
			fp += w;
			len -= w;
			off += w;
		} while( len > 0 );
	}
}
