/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import jcifs.netbios.NbtAddress;
import jcifs.util.DES;
import jcifs.util.MD4;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.net.InetAddress;

class SmbSession {

    // KGS!@#$%
    static final byte[] S8 = {
		(byte)0x4b, (byte)0x47, (byte)0x53, (byte)0x21,
		(byte)0x40, (byte)0x23, (byte)0x24, (byte)0x25
	};
	static void E( byte[] key, byte[] data, byte[] e ) {
		byte[] key7 = new byte[7];
		byte[] e8 = new byte[8];

		for( int i = 0; i < key.length / 7; i++ ) {
			System.arraycopy( key, i * 7, key7, 0, 7 );
			DES des = new DES( key7 );
			des.encrypt( data, e8 );
			System.arraycopy( e8, 0, e, i * 8, 8 );
		}
	}
	static byte[] getPreNTLMResponse( String password, byte[] challenge ) {
		byte[] p14 = new byte[14];
        byte[] p21 = new byte[21];
        byte[] p24 = new byte[24];

        System.arraycopy( password.toUpperCase().getBytes(), 0, p14, 0, password.length() );
        E( p14, S8, p21);
		E( p21, challenge, p24);
        return p24;
	}
	static byte[] getNTLMResponse( String password, byte[] challenge ) {
		byte[] uni = null;
		byte[] p21 = new byte[21];
		byte[] p24 = new byte[24];

		try {
 			uni = password.getBytes( "UnicodeLittleUnmarked" );
		} catch( UnsupportedEncodingException uee ) {
			Log.printStackTrace( "password encryption exception", uee );
		}
		MD4 md4 = new MD4();
		md4.update( uni );
		System.arraycopy( md4.digest(), 0, p21, 0, 16 );
		E( p21, challenge, p24 );
		return p24;
	}

	int uid;
	String username, password, domain;
	SmbTransport transport;
	Vector trees;
	boolean sessionSetup;

	SmbSession( SmbTransport transport, String username,
								String password, String domain ) {
		this.transport = transport;
		this.username = username.toUpperCase();
		this.password = password;
		this.domain = domain.toUpperCase();
		trees = new Vector();
		sessionSetup = false;
	}

	synchronized SmbTree getSmbTree( String path, String service ) {
		SmbTree t;
		for( Enumeration e = trees.elements(); e.hasMoreElements(); ) {
			t = (SmbTree)e.nextElement();
			if( t.matches( path, service )) {
				return t;
			}
		}
		t = new SmbTree( this, path, service );
		trees.addElement( t );
		return t;
	}
	boolean matches( String username, String password, String domain ) {
		return this.username.equals( username.toUpperCase() ) &&
				this.password.equals( password ) &&
				this.domain.equals( domain.toUpperCase() );
	}
	synchronized void sendTransaction( SmbComTransaction request,
								SmbComTransactionResponse response ) throws IOException {

		if( sessionSetup == false ) {
			sessionSetup( null );
		}
		request.uid = uid;
		transport.sendTransaction( request, response );
	}
	synchronized void send( ServerMessageBlock request,
								ServerMessageBlock response ) throws IOException {
		if( sessionSetup == false ) {
			sessionSetup( null );
		}
		request.uid = uid;
		transport.send( request, response );
	}
	synchronized ServerMessageBlock sessionSetup( ServerMessageBlock andx )
								throws IOException {

		if( sessionSetup ) {
			return null;
		}

		Log.println( Log.WARNINGS, "smb session setup warning",
					" requesting session with accountName=" + username +
					",accountPassword=" + password +
					",primaryDomain=" + domain );

		/*
		 * Session Setup And X Request / Response
		 */

		SmbComSessionSetupAndXResponse response = new SmbComSessionSetupAndXResponse();
		transport.send( new SmbComSessionSetupAndX( this, andx ), response );

		uid = response.uid;
		sessionSetup = true;
		return response.andx;
	}
	synchronized void logoff() throws IOException {
		if( sessionSetup == false ) {
			return;
		}

		for( Enumeration e = trees.elements(); e.hasMoreElements(); ) {
			SmbTree t = (SmbTree)e.nextElement();
			t.treeDisconnect();
		}
		sessionSetup = false;

		if( transport.server.security == ServerMessageBlock.SECURITY_SHARE ) {
			return;
		}

		/*
		 * Logoff And X Request / Response
		 */

		SmbComLogoffAndX request = new SmbComLogoffAndX( null );
		request.uid = uid;
		transport.send( request, null );
	}
	public String toString() {
		return "SmbSession[accountName=" + username +
				",accountPassword=" + password +
				",primaryDomain=" + domain +
				",uid=" + uid + "]";
	}
}
