/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import jcifs.netbios.NbtSocket;
import jcifs.netbios.NbtAddress;
import jcifs.util.Config;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PushbackInputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;

class SmbTransport implements Runnable {

	private static final int DEFAULT_MAX_MPX_COUNT = 10;
	private static final int DEFAULT_RESPONSE_TIMEOUT = 10000;
	private static final int DEFAULT_SO_TIMEOUT    = 15000;
	private static final int PUSHBACK_BUF_SIZE     = 64;
	private static final int DEFAULT_RCV_BUF_SIZE  = 4096;
	private static final int DEFAULT_SND_BUF_SIZE  = 1300;

	static final int MID_OFFSET      = 30;
	static final int HEADER_LENGTH   = 32;
	static final int FLAGS_OFFSET    = 9;
	static final int FLAGS_RESPONSE  = 0x80;

	private byte[] rcv_buf, snd_buf;

	private NbtSocket socket;
	private InputStream in;
	private OutputStream out;
	private int localPort, soTimeout, responseTimeout;
	private InetAddress localAddr;
	private Hashtable responseTable;
	private Thread thread;

	private static Vector connections = new Vector();
	private static MpxControl mpxCtrl = new MpxControl();

	Vector sessions;
	boolean useUnicode;

	NbtAddress address;
	int port, rcv_buf_size, snd_buf_size;

	int negotiatedDialectIndex;
	int negotiatedMaxMpxCount;
	int negotiatedMaxBufferSize;
	int negotiatedCapabilities;

	ClientProperties client = new ClientProperties();
	ServerProperties server = new ServerProperties();

	class ClientProperties {
		int maxMpxCount = Config.getInt( "jcifs.smb.client.maxMpxCount",
													DEFAULT_MAX_MPX_COUNT );
		int maxBufferSize = Config.getInt( "jcifs.smb.client.snd_buf_size",
													DEFAULT_SND_BUF_SIZE );
		int sessionKey = 0x00000000;
		/* NT 4 Workstation client capabilities 0x00D4
		 * we don't do NT Status Codes or Level II oplocks
		 * but we could do raw and mpx
		 */
		int flags2 = Config.getInt( "jcifs.smb.client.flags2",
				ServerMessageBlock.FLAGS2_LONG_FILENAMES |
													ServerMessageBlock.FLAGS2_UNICODE );
		int capabilities = Config.getInt( "jcifs.smb.client.capabilities",
									ServerMessageBlock.CAP_UNICODE );
/*
 * ServerMessageBlock.CAP_UNICODE | ServerMessageBlock.CAP_NT_SMBS );
 * Still having a little trouble with SmbComNTCreateAndX
 */
		String nativeOs = Config.getProperty( "jcifs.smb.client.nativeOs",
				System.getProperty( "os.name" ));
		String nativeLanMan = Config.getProperty( "jcifs.smb.client.nativeLanMan", "foo" );
		int vcNumber = 0;
	}
	class ServerProperties {
		byte flags;
		int flags2;
		int maxMpxCount;
		int maxBufferSize;
		int sessionKey;
		// NT 4 Workstation is 0x43FD
		int capabilities;
		String oemDomainName;
		String primaryDomain;
		String nativeOs;
		String nativeLanMan;

		int securityMode;
		int security;
		boolean encryptedPasswords;
		boolean signaturesEnabled;
		boolean signaturesRequired;
		int maxNumberVcs;
		int maxRawSize;
		long serverTime;
		int serverTimeZone;
		int encryptionKeyLength;
		byte[] encryptionKey;
	}

	static synchronized SmbTransport getSmbTransport( NbtAddress address, int port ) {
		return getSmbTransport( address, port,
							Config.getInetAddress( "jcifs.smb.client.laddr", null ),
							Config.getInt( "jcifs.smb.client.lport", 0 ));
	}
	static synchronized SmbTransport getSmbTransport( NbtAddress address, int port,
									InetAddress localAddr, int localPort ) {
		SmbTransport conn;
		for( Enumeration e = connections.elements(); e.hasMoreElements(); ) {
			conn = (SmbTransport)e.nextElement();
			if( conn.matches( address, port, localAddr, localPort )) {
				return conn;
			}
		}
		conn = new SmbTransport( address, port, localAddr, localPort );
		connections.addElement( conn );
		return conn;
	}

	SmbTransport( NbtAddress address, int port, InetAddress localAddr, int localPort ) {
		this.address = address;
		this.port = port;
		this.localAddr = localAddr;
		this.localPort = localPort;

		useUnicode = Config.getBoolean( "jcifs.smb.client.useUnicode", true );
		if( useUnicode == false ) {
			client.flags2 &= 0xFFFF ^ ServerMessageBlock.FLAGS2_UNICODE;
			client.capabilities &= 0xFFFF ^ ServerMessageBlock.CAP_UNICODE;
		}
/*
		if( Config.getBoolean( "jcifs.smb.client.useNTSmbs", true ) == false ) {
			client.capabilities ^= ServerMessageBlock.CAP_NT_SMBS;
		}
*/
		if( Config.getBoolean( "jcifs.smb.client.useNTSmbs", false ) == true ) {
			client.capabilities |= ServerMessageBlock.CAP_NT_SMBS;
		}

		soTimeout = Config.getInt( "jcifs.smb.client.soTimeout", DEFAULT_SO_TIMEOUT );
		responseTimeout = Config.getInt( "jcifs.smb.client.responseTimeout",
															DEFAULT_RESPONSE_TIMEOUT );
		rcv_buf_size = Config.getInt( "jcifs.smb.client.rcv_buf_size", DEFAULT_RCV_BUF_SIZE );
		snd_buf_size = client.maxBufferSize;
		rcv_buf = new byte[rcv_buf_size];
		snd_buf = new byte[snd_buf_size];
		sessions = new Vector();
		responseTable = new Hashtable( 3 );
	}

	synchronized SmbSession getSmbSession() {
		return getSmbSession( null, null, null );
	}
	synchronized SmbSession getSmbSession( String username, String password, String domain ) {
		if( username == null ) {
			username = Config.getProperty( "jcifs.smb.client.username",
						System.getProperty( "user.name", "guest" ));
		}
		if( password == null ) {
			password = Config.getProperty( "jcifs.smb.client.password", "" );
		}
		if( domain == null ) {
			domain = Config.getProperty( "jcifs.smb.client.domain", "?" );
		}
		SmbSession ssn;
		for( Enumeration e = sessions.elements(); e.hasMoreElements(); ) {
			ssn = (SmbSession)e.nextElement();
			if( ssn.matches( username, password, domain )) {
				return ssn;
			}
		}
		ssn = new SmbSession( this, username, password, domain );
		sessions.addElement( ssn );
		return ssn;
	}
	boolean matches( NbtAddress address, int port, InetAddress localAddr, int localPort ) {
		InetAddress defaultLocal = null;
		try {
			defaultLocal = InetAddress.getLocalHost();
		} catch( UnknownHostException uhe ) {
		}
		int p1 = ( port == 0 || port == 139 ) ? 0 : port;
		int p2 = ( this.port == 0 || this.port == 139 ) ? 0 : this.port;
		InetAddress la1 = localAddr == null ? defaultLocal : localAddr;
		InetAddress la2 = this.localAddr == null ? defaultLocal : this.localAddr;
		return address.equals( this.address ) &&
					p1 == p2 &&
					la1.equals( la2 ) &&
					localPort == this.localPort;
	}
	synchronized void open() throws IOException {
		if( socket != null ) {
			return;
		}
		socket = new NbtSocket( address, port, localAddr, localPort );
		socket.setSoTimeout( soTimeout );
		in = new PushbackInputStream( socket.getInputStream(), PUSHBACK_BUF_SIZE );
		out = socket.getOutputStream();
		thread = new Thread( this );
		thread.start();
	}
	synchronized void close() {
		try {
			if( socket != null ) {
				for( Enumeration e = sessions.elements(); e.hasMoreElements(); ) {
					((SmbSession)e.nextElement()).logoff();
				}
				socket.close();
			}
		} catch( IOException ioe ) {
		}
		in = null;
		out = null;
		socket = null;
		mpxCtrl.clear();
		responseTable.clear();
	}
	public void run() {
		int mid, l;
		ServerMessageBlock response;
		try {
			while( socket != null ) {
				if( in.read( rcv_buf, 0, HEADER_LENGTH ) != HEADER_LENGTH ) {
					break;
				}
				((PushbackInputStream)in).unread( rcv_buf, 0, HEADER_LENGTH );
				if( rcv_buf[0] != (byte)0xFF ||
								rcv_buf[1] != (byte)'S' ||
								rcv_buf[2] != (byte)'M' ||
								rcv_buf[3] != (byte)'B' ) {
					Log.println( Log.WARNINGS, "smb warning",
												"bad smb header, purging session message" );
					in.skip( in.available() );
					continue;
				}
				if(( rcv_buf[FLAGS_OFFSET] & FLAGS_RESPONSE ) == FLAGS_RESPONSE ) {
					mid = ServerMessageBlock.readInt2( rcv_buf, MID_OFFSET );
	
					response = (ServerMessageBlock)responseTable.get( new Integer( mid ));
					if( response == null ) {
						Log.println( Log.WARNINGS, "smb warning",
								"no handler for mid=" + mid + ", purging session message" );
						in.skip( in.available() );
						continue;
					}
					if( response instanceof SmbComTransactionResponse ) {
						((Enumeration)response).nextElement();
					}
					response.useUnicode = useUnicode;

					Log.println( Log.DEBUGGING, "smb debugging",
													" new data read from socket" );
					response.readWireFormat( in, rcv_buf, 0 );

					synchronized( response ) {
						response.received = true;

						Log.printMessageData( "smb received from: " + address, response );
						ServerMessageBlock smb = response;
						while( smb instanceof AndXServerMessageBlock &&
										( smb = ((AndXServerMessageBlock)smb).andx ) != null ) {
							Log.printMessageData( "smb andx data", smb );
						}
						Log.printHexDump( "smb received from: " +
													address, rcv_buf, 0, response.length );

						response.notify();
					}
				} else {
					// it's a request(break oplock)
				}
			}
		} catch( InterruptedIOException iioe ) {
		} catch( IOException ioe ) {
			Log.printStackTrace( "exception reading from socket input", ioe );
		}
		close();
	}
	void send( ServerMessageBlock request,
							ServerMessageBlock response ) throws IOException {
		if( socket == null ) {
			open();
			negotiate();
		}
		request.flags2 = client.flags2;
		request.mid = aquireMid();
		request.useUnicode = useUnicode;

		if( response == null ) {
			synchronized( out ) {
				out.write( snd_buf, 0, request.writeWireFormat( snd_buf, 0 ));
				out.flush();

				Log.printMessageData( "smb sent to: " + address, request );
				ServerMessageBlock smb = request;
				while( smb instanceof AndXServerMessageBlock &&
								( smb = ((AndXServerMessageBlock)smb).andx ) != null ) {
					Log.printMessageData( "smb andx data", smb );
				}
				Log.printHexDump( "smb sent to: " + address, snd_buf, 0, request.length );
			}
		} else {
			response.received = false;
			responseTable.put( new Integer( request.mid ), response );

			synchronized( out ) {
				int n = request.writeWireFormat( snd_buf, 0 );
				out.write( snd_buf, 0, n );
				out.flush();

				Log.printMessageData( "smb sent to: " + address, request );
				ServerMessageBlock smb = request;
				while( smb instanceof AndXServerMessageBlock &&
							( smb = ((AndXServerMessageBlock)smb).andx ) != null ) {
					Log.printMessageData( "smb andx data", smb );
				}
				Log.printHexDump( "smb sent to: " + address, snd_buf, 0, request.length );
			}

			synchronized( response ) {
				if( response.received == false ) {
					try {
						response.wait( responseTimeout );
					} catch( InterruptedException ie ) {
					}
				}
			}
			responseTable.remove( new Integer( request.mid ));
			releaseMid( request.mid );
	
			if( response.received == false ) {
				throw new IOException( "time out waiting for response from server: " + address );
			}
			if( response.errorCode != 0 ) {
				throw new SmbException( response.errorCode );
			}
		}
	}
	void sendTransaction( SmbComTransaction request,
							SmbComTransactionResponse response ) throws IOException {
		if( socket == null ) {
			open();
			negotiate();
		}

		/*
		 * this is basically wrong for multi-part
		 * transactions. Trans2FindXxx2 with NT would benefit from
		 * making multi-part responses work(as is I have observed
		 * Trans2FindXxx2 list 4100 files in 8sec and 6500 files
		 * in 12sec).
		 */

		request.flags2 = client.flags2;
		request.mid = aquireMid();
		request.useUnicode = useUnicode;
		request.maxBufferSize = negotiatedMaxBufferSize;
		response.received = false;
		responseTable.put( new Integer( request.mid ), response );

		while( request.hasMoreElements() ) {
			request.nextElement();
			synchronized( out ) {
				out.write( snd_buf, 0, request.writeWireFormat( snd_buf, 0 ));
				out.flush();

				Log.printMessageData( "smb sent to: " + address, request );
				Log.printHexDump( "smb sent to: " + address, snd_buf, 0, request.length );
			}
			synchronized( response ) {
				if( response.received == false ) {
					try {
						response.wait( responseTimeout );
					} catch( InterruptedException ie ) {
					}
				}
			}
		}
		responseTable.remove( new Integer( request.mid ));
		releaseMid( request.mid );

		if( response.received == false ) {
			throw new IOException( "time out waiting for response from server" );
		}
		if( response.errorCode != 0 ) {
			throw new SmbException( response.errorCode );
		}
	}
	synchronized void negotiate() throws IOException {

		Log.println( Log.WARNINGS, "smb negotiation warning",
							" requesting negotiation with " + address );

		/*
		 * Negotiate Protocol Request / Response
		 */

		SmbComNegotiateResponse response = new SmbComNegotiateResponse();
		send( new SmbComNegotiate(), response );

		if( response.dialectIndex > 10 ) {
			throw new IOException( "smb failed to negotiate dialect" );
		}

		server.securityMode        = response.securityMode;
		server.security            = response.security;
		server.encryptedPasswords  = response.encryptedPasswords;
		server.signaturesEnabled   = response.signaturesEnabled;
		server.signaturesRequired  = response.signaturesRequired;
		negotiatedDialectIndex     = response.dialectIndex;;
		server.maxMpxCount         = response.maxMpxCount;
		negotiatedMaxMpxCount      = client.maxMpxCount < server.maxMpxCount ?
										client.maxMpxCount : server.maxMpxCount;
		mpxCtrl.maxMpxCount        = negotiatedMaxMpxCount < 1 ?
										1 : negotiatedMaxMpxCount;
		server.maxNumberVcs        = response.maxNumberVcs;
		server.maxBufferSize       = response.maxBufferSize;
		negotiatedMaxBufferSize    = client.maxBufferSize < server.maxBufferSize ?
										client.maxBufferSize : server.maxBufferSize;
		server.maxRawSize          = response.maxRawSize;
		server.sessionKey          = response.sessionKey;
		server.capabilities        = response.capabilities;
		negotiatedCapabilities     = client.capabilities & server.capabilities;
		if(( negotiatedCapabilities & ServerMessageBlock.CAP_UNICODE ) == 0 ) {
			// server doesn't want unicode
			if( Config.getBoolean( "jcifs.smb.client.useUnicode", false )) {
				// force unicode
				negotiatedCapabilities |= ServerMessageBlock.CAP_UNICODE;
			} else {
				// not explicitly set to true so flip unicode off as server requests
				useUnicode = false;
				client.flags2 &= 0xFFFF ^ ServerMessageBlock.FLAGS2_UNICODE;
			}
		}
		server.serverTime          = response.serverTime;
		server.serverTimeZone      = response.serverTimeZone;
		server.encryptionKeyLength = response.encryptionKeyLength;
		server.encryptionKey       = response.encryptionKey;
		server.oemDomainName       = response.oemDomainName;
	}
	public String toString() {
		return "SmbTransport[address=" + address +
			",port=" + socket.getPort() +
			",localAddr=" + socket.getLocalAddress() +
			",localPort=" + socket.getLocalPort() + "]";
	}

	static int aquireMid() throws IOException {
		try {
			return mpxCtrl.aquireMid();
		} catch( InterruptedException ie ) {
			throw new IOException( ie.getMessage() );
		}
	}
	static void releaseMid( int mid ) {
		mpxCtrl.releaseMid( mid );
	}

    static class MpxControl {
		MpxListNode first, last;
        class MpxListNode {
            int i;  
            MpxListNode next;
			MpxListNode( int id ) {
				i = id;
				next = null;
			}
        }       
		int nextMid, mpxCount, maxMpxCount;

		MpxControl() {
			nextMid = 0;
			mpxCount = 0;
			maxMpxCount = 1;
			first = last = null;
		}

		synchronized void clear() {
			nextMid = 0;
			mpxCount = 0;
			maxMpxCount = 1;
			first = last = null;
		}
		synchronized int aquireMid() throws InterruptedException {
			while( mpxCount >= maxMpxCount ) {
				wait();
			}
			if(( ++nextMid % 0xFFFF ) == 0 ) {
				nextMid++;
			}
			add( nextMid );
			mpxCount++;
			return nextMid;
		}
		synchronized void releaseMid( int i ) {
			remove( i );
			mpxCount--;
			notify();
		}
        synchronized boolean contains( int i ) {
			for( MpxListNode tmp = first; tmp != null; tmp = tmp.next ) {
				if( tmp.i == i ) {
					return true;
				}
			}
			return false;
        }       
        synchronized void add( int i ) {
			if( contains( i )) {
				return;
			}
			if( first == null ) {
				first = last = new MpxListNode( i );
			} else {
				last = last.next = new MpxListNode( i );
			}
        }
        synchronized void remove( int i ) {
			if( first == null ) {
				return;
			} else if( first == last && first.i == i ) {
				first = last = null;
				return;
			}
			MpxListNode tmp, prev;
			for( tmp = prev = first; tmp != null; tmp = tmp.next ) {
				if( tmp.i == i ) {
					if( tmp == first ) {
						first = first.next;
					}
					if( tmp == last ) {
						last = prev;
						last.next = null;
					}
				}
				prev = tmp;
			}
        }       
    }
}
