/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import java.io.IOException;

class SmbTree {

	int tid;
	String path;
	String service = "?????";
	SmbSession session;
	boolean treeConnected = false;

	SmbTree( SmbSession session, String path, String service ) {
		this.session = session;
		this.path = path;
		if( service != null && service.startsWith( "??" ) == false ) {
			this.service = service;
		}
	}

	boolean matches( String path, String service ) {
		return this.path.equals( path ) &&
				( service == null || service.startsWith( "??" ) ||
				this.service.equals( service ));
	}
	synchronized void sendTransaction( SmbComTransaction request,
								SmbComTransactionResponse response ) throws IOException {
		if( treeConnected == false ) {
			treeConnect( null );
		}
		request.tid = tid;
		session.sendTransaction( request, response );
	}
	synchronized void send( ServerMessageBlock request,
								ServerMessageBlock response ) throws IOException {
		if( treeConnected == false ) {
			treeConnect( null );
		}
		request.tid = tid;
		session.send( request, response );
	}
	synchronized ServerMessageBlock treeConnect( ServerMessageBlock andx )
								throws IOException {

		if( treeConnected ) {
			return null;
		}

		Log.println( Log.WARNINGS, "smb tree connect warning",
				" requesting tree connect with path=" + path +
				",service=" + service );

		/*
		 * Tree Connect And X Request / Response
		 */

		SmbComTreeConnectAndXResponse response = new SmbComTreeConnectAndXResponse();
		SmbComTreeConnectAndX request = new SmbComTreeConnectAndX( session, path,
														service, andx );
		session.send( request, response );

		tid = response.tid;
		service = response.service;
		treeConnected = true;
		return response.andx;
	}
	synchronized void treeDisconnect() {
		if( treeConnected == false ) {
			return;
		}
		try {
			send( new SmbComTreeDisconnect(), null );
		} catch( IOException ioe ) {
			Log.printStackTrace( "smb tree disconnect exception", ioe );
		} finally {
			treeConnected = false;
		}
	}
	boolean checkDirectory( String path ) throws IOException {

		Log.println( Log.WARNINGS, "smb check directory warning",
				" checking directory with path=" + path );

		/*
		 * Check Directory Request / Response
		 */

		try {
			send( new SmbComCheckDirectory( path ), new SmbComBlankResponse() );
		} catch( SmbException se ) {
			if( se.errorClass == SmbException.ERRDOS &&
									se.errorCode == SmbException.ERRbadpath ) {
				return false;
			} else {
				throw se;
			}
		}
		return true;
	}
	Info queryPath( String path ) throws IOException {

		Log.println( Log.WARNINGS, "smb query path warning",
				" querying path=" + path );

		// how do I know if Trans2Query is supported?

		Info result = null;
		if(( session.transport.negotiatedCapabilities &
									ServerMessageBlock.CAP_NT_SMBS ) ==
									ServerMessageBlock.CAP_NT_SMBS ) {
			/*
			 * Trans2 Query Path Information Request / Response
			 */

			int infoLevel = Trans2QueryPathInformationResponse.SMB_QUERY_FILE_BASIC_INFO;

			Trans2QueryPathInformationResponse response =
									new Trans2QueryPathInformationResponse( infoLevel );
			try {
				sendTransaction( new Trans2QueryPathInformation( path, infoLevel ), response );
			} catch( SmbException se ) {
				if( se.errorClass == SmbException.ERRDOS &&
										se.errorCode == SmbException.ERRbadfile ) {
					return null;
				} else {
					throw se;
				}
			}
			result = response.info;
		} else {

			/*
			 * Query Information Request / Response
			 */

			SmbComQueryInformationResponse response = new SmbComQueryInformationResponse();
			try {
				send( new SmbComQueryInformation( path ), response );
			} catch( SmbException se ) {
				if( se.errorClass == SmbException.ERRDOS &&
										se.errorCode == SmbException.ERRbadfile ) {
					return null;
				} else {
					throw se;
				}
			}
			result = response;
		}
		return result;
	}
	String[] find( String path ) throws IOException {
		int sid, count, i, j;
		String[] results;
		String filename;

		if( path.equals( "\\" )) {
			path = "\\*";
		} else {
			path += "\\*";
		}

		Log.println( Log.WARNINGS, "smb find warning",
				" find with path=" + path );

		Trans2FindFirst2Response response = new Trans2FindFirst2Response();
		try {
			sendTransaction( new Trans2FindFirst2( path ), response );
		} catch( SmbException se ) {
			if( se.errorClass == SmbException.ERRDOS &&
							se.errorCode == SmbException.ERRbadfile ) {
				return null;
			}
			throw se;
		}
		sid = response.sid;
		count = response.searchCount;
		j = 0;

		results = new String[Math.max( 10, count )];
		/*
		 * using hashCode is much more efficient than
		 * filename.equals( "." ) || filename.equals( ".." )
		 */
		int h1 = new String( "." ).hashCode();
		int h2 = new String( ".." ).hashCode();
		i = 0;
		while( j < count ) {
			filename = response.results[i++].filename;
			if( filename.length() < 3 ) {
				int h = filename.hashCode();
				if( h == h1 || h == h2 ) {
					count--;
					continue;
				}
			}
			results[j++] = filename;
		}

		// only difference between first2 and next2 responses is subCommand so let's recycle
		response.subCommand = SmbComTransaction.TRANS2_FIND_NEXT2;

		while( response.isEndOfSearch == false && response.searchCount > 0 ) {
			sendTransaction( new Trans2FindNext2( sid, response.lastName ), response );
			count += response.searchCount;

			if( count > results.length ) {
				String[] tmp = results;
				results = new String[Math.max( results.length * 2, count )];
				System.arraycopy( tmp, 0, results, 0, j );
			}
			i = 0;
			while( j < count ) {
				filename = response.results[i++].filename;
				if( filename.length() < 3 ) {
					int h = filename.hashCode();
					if( h == h1 || h == h2 ) {
						count--;
						continue;
					}
				}
				results[j++] = filename;
			}
		}

		try {
			send( new SmbComFindClose2( sid ), new SmbComBlankResponse() );
		} catch( SmbException se ) {
			Log.printStackTrace( "smb find close response exception", se );
		}

		if( results.length != count ) {
			String[] tmp = results;
			results = new String[count];
			System.arraycopy( tmp, 0, results, 0, count );
		}

		return results;
	}
	boolean rename( String oldFileName, String newFileName ) throws IOException {

		Log.println( Log.WARNINGS, "smb rename warning",
				" oldFileName=" + oldFileName +
				",newFileName=" + newFileName );

		/*
		 * Rename Request / Response
		 */

		SmbComBlankResponse response = new SmbComBlankResponse();
		try {
			send( new SmbComRename( oldFileName, newFileName ), response );
		} catch( SmbException se ) {
			if( se.errorClass == SmbException.ERRDOS &&
									( se.errorCode == SmbException.ERRbadpath ||
									se.errorCode == SmbException.ERRbadfile )) {
				return false;
			} else {
				throw se;
			}
		}
		return true;
	}
	boolean copy( String sourceFileName, String targetFileName, int tid2 )
													throws IOException {

		Log.println( Log.WARNINGS, "smb copy warning",
				" sourceFileName=" + sourceFileName +
				",targetFileName=" + targetFileName );

		/*
		 * Copy Request / Response
		 */

		SmbComCopyResponse response = new SmbComCopyResponse();
		try {
			send( new SmbComCopy( sourceFileName, targetFileName, tid2 ), response );
		} catch( SmbException se ) {
			if( se.errorClass == SmbException.ERRDOS &&
									( se.errorCode == SmbException.ERRbadpath ||
									se.errorCode == SmbException.ERRbadfile )) {
				return false;
			} else {
				throw se;
			}
		}
		return true;
	}
	boolean delete( String fileName, boolean isDirectory ) throws IOException {

		Log.println( Log.WARNINGS, "smb delete warning",
				" fileName=" + fileName );

		/*
		 * Delete or Delete Directory Request / Response
		 */

		ServerMessageBlock request;
		if( isDirectory ) {

			/* Recursively delete directory contents using post order DFS
			 */

			String[] list = find( fileName );
			for( int i = 0; i < list.length; i++ ) {
				String p = fileName + "\\" + list[i];
				Info info = queryPath( p );
				boolean dir = ( info.getAttributes() & SmbFile.ATTR_DIRECTORY ) ==
														SmbFile.ATTR_DIRECTORY;
				delete( p, dir );
			}

			request = new SmbComDeleteDirectory( fileName );
		} else {
			request = new SmbComDelete( fileName );
		}

		try {
			send( request, new SmbComBlankResponse() );
		} catch( SmbException se ) {
			if( se.errorClass == SmbException.ERRDOS &&
									( se.errorCode == SmbException.ERRbadpath ||
									se.errorCode == SmbException.ERRbadfile )) {
				return false;
			} else {
				throw se;
			}
		}
		return true;
	}
	int open( String name, int flags ) throws IOException {

		Log.println( Log.WARNINGS, "smb open warning", " name=" + name );

		/*
		 * Open AndX Request / Response
		 */

		/* To query server capabilities we must ensure the negotiation has talken place
		 */

		if( treeConnected == false ) {
			treeConnect( null );
		}

		if(( session.transport.negotiatedCapabilities &
									ServerMessageBlock.CAP_NT_SMBS ) ==
									ServerMessageBlock.CAP_NT_SMBS ) {
			// NT Smb's are disabled at the moment
			SmbComNTCreateAndXResponse response = new SmbComNTCreateAndXResponse();
			send( new SmbComNTCreateAndX( name, null ), response );
			return response.fid;
		} else {
			SmbComOpenAndXResponse response = new SmbComOpenAndXResponse();
			send( new SmbComOpenAndX( name, flags, null ), response );
			return response.fid;
		}
	}
	boolean createDirectory( String directoryName ) throws IOException {

		Log.println( Log.WARNINGS, "smb create directory warning",
				" directoryName=" + directoryName );

		/*
		 * Create Directory Request / Response
		 */

		try {
			send( new SmbComCreateDirectory( directoryName ), new SmbComBlankResponse() );
		} catch( SmbException se ) {
			if( se.errorClass == SmbException.ERRDOS &&
									se.errorCode == SmbException.ERRnoaccess ) {
				return false;
			} else {
				throw se;
			}
		}
		return true;
	}
	public String toString() {
		return "SmbTree[path=" + path +
			",service=" + service +
			",tid=" + tid + "]";
	}
}
