/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.util;

import java.util.Hashtable;
import java.util.Properties;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * This class uses a static {@link jcifs.util.PropertiesTree} to act as a
 * cental repository of configuration information. When initialized for the
 * first time the system properties will constitute it's initial properties.
 * <p>
 * See also <a href="../smb/package-summary.html">Setting JCIFS Properties</a>
 */

public class Config {

	private static PropertiesTree pt = new PropertiesTree();
	private static Hashtable alreadyLoaded = new Hashtable();

	static {
		Properties sys = System.getProperties();
		synchronized( sys ) {
			String pkgs = sys.getProperty( "java.protocol.handler.pkgs" );
			if( pkgs == null ) {
				pkgs = "jcifs";
			} else {
				pkgs += "|jcifs";
			}
			sys.put( "java.protocol.handler.pkgs", pkgs );
		}
		StringBuffer add = new StringBuffer();
		StringBuffer ign = new StringBuffer();
		for( Enumeration e = sys.propertyNames(); e.hasMoreElements(); ) {
			String key = (String)e.nextElement();
			String val = (String)sys.get( key );
			try {
				pt.put( key, val );
				add.append( " " + key + "=" + val + Log.NL );
			} catch( IllegalArgumentException iae ) {
				ign.append( " " + key + "=" + val + Log.NL );
			}
		}

		// This could be better. More dynamic.

		String mask = Config.getProperty( "jcifs.util.log" );
		if( mask != null ) {
			StringTokenizer st = new StringTokenizer( mask, "," );
			while( st.hasMoreTokens() ) {
				String s = st.nextToken().toUpperCase();
				if( s.startsWith( "ALL" )) {
					Log.addMask( Log.ALL );
				} else if( s.startsWith( "EXC" )) {
					Log.addMask( Log.EXCEPTIONS );
				} else if( s.startsWith( "WAR" )) {
					Log.addMask( Log.WARNINGS );
				} else if( s.startsWith( "DEB" )) {
					Log.addMask( Log.DEBUGGING );
				} else if( s.startsWith( "HEX" )) {
					Log.addMask( Log.HEX_DUMPS );
				} else if( s.startsWith( "NON" )) {
					Log.setMask( Log.NONE );
					break;
				}
			}
		}
		Log.println( Log.WARNINGS, "added properties", add.toString() );
		Log.println( Log.WARNINGS, "ignoring properties", ign.toString() );

		load( "jcifs" ); // ensure jcifs.properties file is loaded
	}

	private static void load( String key ) {
		if( alreadyLoaded.get( key ) != null ) {
			return;
		} else {
			alreadyLoaded.put( key, key );
		}
		String filename = System.getProperty( key + ".properties" );
		FileInputStream in = null;
		try {
			if( filename != null ) {
		 		in = new FileInputStream( filename );
			}
		} catch( FileNotFoundException fnfe ) {
			Log.printStackTrace( "config exception", fnfe );
		}
		if( in != null ) {
			try {
				pt.load( in );
				in.close();
			} catch( IOException ioe ) {
				Log.printStackTrace( "config load exception", ioe );
			}
		} else {
			int i = key.lastIndexOf( "." );
			if( i > 0 ) {
				key = key.substring( 0, i );
				load( key );
			}
		}
	}

	/**
	 * This method is actually empty. Indirectly it ensures that
	 * the static initializer is called which loads the system
	 * properties. It also adds the jcifs package name to the System
	 * properties for the java.net.URL handler.
	 */

	public static void init() {};

	/**
	 * Put a property into the static PropertiesTree. This is probably
	 * the easiest way to set jcifs properties.
	 */

	public static Object setProperty( String key, String value ) {
		return pt.put( key, value );
	}

	/**
	 * Retrieve a property from the static PropertiesTree.
	 */

	public static Object get( String key ) {
		Object obj = pt.get( key );
		if( obj == null ) {
			load( key );
			obj = pt.get( key );
		}
		return obj;
	}

	/**
	 * Retrieve a property from the static PropertiesTree. The
	 * default value will be returned if the key is not found in the
	 * PropertiesTree.
	 */

	public static String getProperty( String key, String def ) {
		Object obj = Config.get( key );
		return obj == null ? def : (String)obj;
	}

	/**
	 * Retrieve a property from the static PropertiesTree.
	 */

	public static String getProperty( String key ) {
		return (String)Config.get( key );
	}
	public static int getInt( String key, int def ) {
		Object obj = Config.get( key );
		if( obj != null ) {
			try {
				def = Integer.parseInt( (String)obj );
			} catch( NumberFormatException nfe ) {
				Log.printStackTrace( "configuration warning", nfe );
			}
		}
		return def;
	}

	/**
	 * Retrieve an int from the PropertiesTree.
	 */

	public static int getInt( String key ) {
		int result = -1;
		Object obj = Config.get( key );
		if( obj != null ) {
			try {
				result = Integer.parseInt( (String)obj );
			} catch( NumberFormatException nfe ) {
				Log.printStackTrace( "configuration warning", nfe );
			}
		}
		return result;
	}

	/** 
	 * Retrieve an address from the PropertiesTree and return as InetAddress.
	 */

	public static InetAddress getInetAddress( String key, InetAddress def ) {
		String addr = getProperty( key );
		if( addr != null ) {
			try {
				def = InetAddress.getByName( addr );
			} catch( UnknownHostException uhe ) {
			}
		}
		return def;
	}

	/**
	 * Retrieve a boolean value.
	 */

	public static boolean getBoolean( String key, boolean def ) {
		String b = getProperty( key );
		if( b != null ) {
			def = b.equals( "true" );
		}
		return def;
	}
}
