/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 *                     "Christopher R. Hertel" <crh@nts.umn.edu>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.util;

import java.io.Writer;
import java.io.OutputStream;

/**
 * This log utility uses a combination of bit mask control and io stream
 * functionality. All methods check the mask with the mask passed as a
 * parameter to see if any further work(logging) should be performed. This
 * provides explicit control at runtime over what is logged. The io stream
 * model allows the {@link java.io.PrintWriter} to be set and provides
 * many of the common {@link java.io.PrintWriter} methods that are usefull
 * for streams. A little overkill but it is easy to use, implement,
 * understand, and better than the plain Singlton style utilities.
 * 
 * <p> Generally you need only be concerned with setting the mask with the
 * desired class members of this Log class or any other class that extend
 * this one like so:
 * 
 * <p><blockquote><pre>
 * Log.setMask( jcifs.util.Log.EXCEPTIONS +
 *     jcifs.util.Log.WARNINGS +
 *     any.other.package.Log.DEBUGGING );
 * </pre></blockquote>
 * <p> I guess the convention is to provide a class for each package
 * called <code>Log</code> that extends this one and handles package
 * specific messages. The only problem with this technique is that new
 * <code>Log</code> classes must be carefull to use bits of the mask that
 * are not already in use.
 *
 * @version   0.1, 00/03/13
 * @author    Michael B. Allen
 * @see       jcifs.netbios.Log
 * @since     jcifs-0.1
 */

public class Log {

    // supress javadoc constructor summary by
    // removing 'protected' while javadocing
    protected Log() {}

    private static final char[] hexChars = { 
        '0', '1', '2', '3', '4', '5',
        '6', '7', '8', '9', 'A', 'B',
        'C', 'D', 'E', 'F'
    };
    private static final char[] spaceChars = {
        ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ',
        ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ',
        ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '
    };

/**
 * The systems line separator.
 */

    public static final String NL = System.getProperty( "line.separator" );

/**
 * The systems line separator length.
 */

    public static final int NL_LENGTH = NL.length();

/**
 * Mask to indicate that no messages should be logged(not even exceptions).
 */

    public static final int NONE        = 0x00000000;

/**
 * Log all messages.
 */

    public static final int ALL            = 0xFFFFFFFF;

/** 
 * This is the default mask. However, be aware that specifying a
 * mask without adding this mask will result in exceptions not being
 * logged. One should always add this EXCEPTIONS mask to the {@link
 * jcifs.util.Log#setMask(int mask)} expression unless you are explicitly
 * trying to suppress them. Ofcourse only exceptions that go through
 * the {@link jcifs.util.Log#printStackTrace(String desc, Throwable t )}
 * facility can be filtered out.
 */

    public static final int EXCEPTIONS    = 0x00000001;

/**
 * This is just a standard convention for warning messages.
 */

    public static final int WARNINGS    = 0x00000002;

/** 
 * Use this when your just trying to log information while coding. It's
 * an alternative for the ever popular <code>System.out.println( "MADE
 * IT!" )</code>.
 */

    public static final int DEBUGGING    = 0x00000004;

/** 
 * This is used by the {@link jcifs.util.Log#printHexDump}
 * function provided by this class.
 */ 

    public static final int HEX_DUMPS    = 0x00000008;

/** 
 * The shared output stream shared by all <code>Log</code> classes. The
 * defualt is <code>System.out</code>.
 */

    protected static LogWriter out = new LogWriter( System.out );

/** 
 * This is the integer mask shared by all Log classes. By defualt it is
 * set to EXCEPTIONS.
 */ 

    protected static int mask = EXCEPTIONS;

    // Setting Methods

/** 
 * Set the {@link java.io.Writer} used as the underlying stream. To write
 * the log to a file, simply do
 * <p><blockquote><pre>
 *  setPrintWriter( new FileWriter( "log.txt" ))
 * </pre></blockquote>
 *
 * @param out the <code>Writer</code> for the underlying stream
 */

    public static void setPrintWriter( Writer out ) {
        Log.out = new LogWriter( out );
    }

/** 
 * Specify an {@link java.io.OutputStream} to be used as the underlying
 * stream.
 * 
 * @param out the output stream
 */

    public static void setPrintWriter( OutputStream out ) {
        Log.out = new LogWriter( out );
    }

/** 
 * This controls what is passed to {@link java.text.SimpleDateFormat} to
 * control what the timestamp looks like in the log stream. For example, a
 * format of <code>"EEE, MMM d, h:mm:ss a"</code> would generate timestamps
 * that read <code>Tue, Mar 14, 4:57:02 PM</code>. This is a whistle.
 *
 * @param the format string
 */ 

    public static void setDateFormat( String format ) {
        out.setDateFormat( format );
    }

/** 
 * Set the mask used to screen all calls to logging methods. The idea
 * here is that the mask can be specified as an arithmetic expression
 * to perovide a crude but effective syslog style way to dictate what it
 * logged. For example:
 * <p><blockquote><pre>
 *  setMask( Log.EXCEPTIONS +
 *           Log.HEX_DUMPS +
 *           jcifs.netbios.Log.PACKET_DATA );
 *
 *  setMask( Log.ALL - jcifs.netbios.Log.PACKET_DIAGRAMS );
 * </pre></blockquote>
 */ 

    public static void setMask( int mask ) {
        Log.mask = mask;
    }
	public static boolean isSet( int mask ) {
		return ( Log.mask & mask ) == mask;
	}
	public static void addMask( int mask ) {
		Log.mask |= mask;
	}

/** 
 * Provides standard way to log any {@link java.lang.Throwable} object like
 * an {@link java.lang.Exception}.
 */
 
      public static void printStackTrace( String desc, Throwable t ) {
        if(( EXCEPTIONS & mask ) != 0 )
            out.printStackTrace( desc, t );
    }

    // Output Methods Used Only By This Package
/** 
 * Hex dumps are ubiquitous enough to provide a standard and easy-to-use
 * method for logging them. All that is required to use this is setting
 * the mask for {@link jcifs.util.Log#HEX_DUMPS}. The hex dump output is
 * represended in the standard form and is quite efficient. Many mega bytes
 * of source data could be dumped to a file without worry.
 * 
 * <p><blockquote><pre>
 * 00000: 04 d2 29 00 00 01 00 00 00 00 00 01 20 45 47 46  |..)......... EGF|
 * 00010: 43 45 46 45 45 43 41 43 41 43 41 43 41 43 41 43  |CEFEECACACACACAC|
 * 00020: 41 43 41 43 41 43 41 43 41 43 41 41 44 00 00 20  |ACACACACACAAD.. |
 * 00030: 00 01 c0 0c 00 20 00 01 00 00 00 00 00 06 20 00  |..... ........ .|
 * 00040: ac 22 22 e1                                      |."".            |
 * </blockquote></pre>
 */

	public static void printHexDump( String desc, byte[] src ) {
		printHexDump( desc, src, 0, src.length );
	}
    public static void printHexDump( String desc, byte[] src, int srcIndex, int length ) {
        if(( HEX_DUMPS & mask ) == 0 || length == 0 ) {
            return;
        }
        int s = length % 16;
        int r = ( s == 0 ) ? length / 16 : length / 16 + 1;
        char[] c = new char[r * (74 + NL_LENGTH)];
        char[] d = new char[16];
        int i;
        int si = 0;
        int ci = 0;
        do {
            toHexChars( si, c, ci, 5 );
            ci += 5;
            c[ci++] = ':';
            do {
                if( si == length ) {
                    int n = 16 - s;
                    System.arraycopy( spaceChars, 0, c, ci, n * 3 );
                    ci += n * 3;
                    System.arraycopy( spaceChars, 0, d, s, n );
                    break;
                }
                c[ci++] = ' ';
                i = src[srcIndex + si] & 0xFF;
                toHexChars( i, c, ci, 2 );
                ci += 2; 
                if( i < 0 || Character.isISOControl( (char)i )) {
                    d[si % 16] = '.';
                } else {
                    d[si % 16] = (char)i;
                }
            } while(( ++si % 16 ) != 0 );
            c[ci++] = ' ';
            c[ci++] = ' ';
            c[ci++] = '|';
            System.arraycopy( d, 0, c, ci, 16 );
            ci += 16;
            c[ci++] = '|';
            NL.getChars( 0, NL_LENGTH, c, ci );
            ci += NL_LENGTH;
        } while( si < length );
        out.println( desc, c );
    }

/** 
 * This is an alternative to the <code>java.lang.Integer.toHexString</cod>
 * method. It is an efficient relative that also will pad the left side so
 * that the result is <code>size</code> digits.
 */ 

    public static String getHexString( int val, int size ) {
        char[] c = new char[size];
        toHexChars( val, c, 0, size );
        return new String( c );
    }
    public static String getHexString( long val, int size ) {
        char[] c = new char[size];
        toHexChars( val, c, 0, size );
        return new String( c );
    }

	public static String getHexString( byte[] src, int srcIndex, int size ) {
		char[] c = new char[size];
		size = ( size % 2 == 0 ) ? size / 2 : size / 2 + 1;
		for( int i = 0, j = 0; i < size; i++ ) {
			c[j++] = hexChars[(src[i] >> 4 ) & 0x0F];
			if( j == c.length ) {
				break;
			}
			c[j++] = hexChars[src[i] & 0x0F];
		}
		return new String( c );
	}

/** 
 * This is the same as {@link jcifs.util.Log#getHexString(int val, int
 * size)} but provides a more practical form when trying to avoid {@link
 * java.lang.String} concatenation and {@link java.lang.StringBuffer}.
 */ 

    public static void toHexChars( int val, char dst[], int dstIndex, int size ) {
        while( size > 0 ) {
            dst[dstIndex + size - 1] = hexChars[val & 0x000F];
            if( val != 0 ) {
                val >>>= 4;
            }
            size--;
        }
    }
    public static void toHexChars( long val, char dst[], int dstIndex, int size ) {
        while( size > 0 ) {
            dst[dstIndex + size - 1] = hexChars[(int)( val & 0x000FL )];
            if( val != 0 ) {
                val >>>= 4;
            }
            size--;
        }
    }

/** 
 * Print an int. The bits of the integer <code>type</code> must match the
 * {@link jcifs.util.Log#mask}.
 */

    public static void println( int type, String desc, int x ) {
        if(( type & mask ) != 0 )
            out.println( desc, x );
    }

/** 
 * Print a char[] array. The bits of the integer <code>type</code> must
 * match the {@link jcifs.util.Log#mask}.
 */
 
    public static void println( int type, String desc, char x[] ) {
        if(( type & mask ) != 0 )
            out.println( desc, x );
    }

/** 
 * Print a {@link java.lang.String}. The bits of the integer
 * <code>type</code> must match the {@link jcifs.util.Log#mask}.
 */
 
    public static void println( int type, String desc, String x ) {
        if(( type & mask ) != 0 )
            out.println( desc, x );
    }

/** 
 * Print an {@link java.lang.Object}. The bits of the integer
 * <code>type</code> must match the {@link jcifs.util.Log#mask}.
 */
 
    public static void println( int type, String desc, Object x ) {
        if(( type & mask ) != 0 )
            out.println( desc, x );
    }
}

