/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.util;

import java.io.*;
import java.util.GregorianCalendar;
import java.util.Date;
import java.text.SimpleDateFormat;

/** 
 * This class is basically a {@link java.io.PrintWriter} that prepends
 * a timestamp before each message is written to the underlying
 * stream. The methods for printing are a pertainent subset from {@link
 * java.io.PrintWriter}.
 *
 * @see jcifs.util.Log
 */
 
public class LogWriter {

    private Writer out;
    private PrintWriter eout;
    private boolean trouble = false;
    private String lineSeparator;

/**
 * The object used to synchronize operations on this stream
 */

    public Object lock;

/** 
 * Create a new <code>LogWriter</code> instance.
 */
 
    public LogWriter( Writer out ) {
        this.lock = this;
        this.out = out;
        eout = new PrintWriter( out );
        lineSeparator = System.getProperty( "line.separator" );
    }

/** 
 * Create a new <code>LogWriter</code> instance that uses the specified
 * output stream as it's underlying stream
 */
 
    public LogWriter( OutputStream out ) {
        this( new OutputStreamWriter( out ));
    }

    static String dateString;
    static GregorianCalendar cal;
    static SimpleDateFormat sdf;

    static {
        cal = new GregorianCalendar();
        setDateFormat( "MMM d hh:mm:ss" );
    }

/** 
 * Set the date format used to render the timestamp written with each
 * message.
 * 
 * @param format the format string used
 * @see java.text.SimpleDateFormat
 */
 
    public static void setDateFormat( String format ) {
        sdf = new SimpleDateFormat( format );
        sdf.setCalendar( cal );
    }

    private void ensureOpen() throws IOException {
        if( out == null )
            throw new IOException( "Stream closed" );
    }

/**
 * Close the stream
 */

    public void close() {
        try {
            synchronized( lock ) {
                if( out == null )
                    return;
                out.close();
                out = null;
            }
        } catch( IOException x ) {
            trouble = true;
        }
    }

/**
 * Flush the stream and check error state
 */

    public boolean checkError() {
        if( out != null ) {
            try {
                synchronized( lock ) {
                    ensureOpen();
                    out.flush();
                }
            } catch( IOException x ) {
                trouble = true;
            }
        }
        return trouble;
    }

/** 
 * Indicate that an error has occured
 */

    public void setError() {
        trouble = true;
    }

/**
 * This is a pervasive enough feature for a logging utility that it is
 * built in.
 *
 * @param desc a brief description of the source of the message to follow
 * @param t the <code>Exception</code> to obtain the message to be printed from
 */

    public void printStackTrace( String desc, Throwable t ) {
        try {
            synchronized( lock ) {
                ensureOpen();
                out.write( sdf.format( new Date() ));
                out.write( " - " );
                out.write( desc );
                out.write( lineSeparator );
                t.printStackTrace( eout );
                out.flush();
            }
        } catch( InterruptedIOException x ) {
            Thread.currentThread().interrupt();
        } catch( IOException ioe ) {
            trouble = true;
        }
    }

/** 
 * Print an <code>int</code> to the log stream. Keep in mind that every
 * call to a <code>PrintWriter</code> like method will print a new timestamp.
 * Using several calls together will not concatonate several messages
 * together under one timestamp.
 *
 * @param desc a brief description of the source of the message to follow
 * @param i the <code>int</code> to print
 */
 
    public void println( String desc, int i ) {
        try {
            synchronized( lock ) {
                ensureOpen();
                out.write( sdf.format( new Date() ));
                out.write( " - " );
                out.write( desc );
                out.write( lineSeparator );
                out.write( Integer.toString( i ));
                out.write( lineSeparator );
                out.flush();
            }
        } catch( InterruptedIOException x ) {
            Thread.currentThread().interrupt();
        } catch( IOException ioe ) {
            trouble = true;
        }
    }

/** 
 * Print an array of <code>char</code> to the log.
 * 
 * @param desc a brief description of the source of the message to follow
 * @param c the message
 */
 
    public void println( String desc, char c[] ) {
        try {
            synchronized( lock ) {
                ensureOpen();
                out.write( sdf.format( new Date() ));
                out.write( " - " );
                out.write( desc );
                out.write( lineSeparator );
                out.write( c );
                out.write( lineSeparator );
                out.flush();
            }
        } catch( InterruptedIOException x ) {
            Thread.currentThread().interrupt();
        } catch( IOException ioe ) {
            trouble = true;
        }
    }

/** 
 * Print a <code>String</code> to the log.
 * 
 * @param desc a brief description of the source of the message to follow
 * @param s the string message
 */ 

    public void println( String desc, String s ) {
        try {
            synchronized( lock ) {
                ensureOpen();
                out.write( sdf.format( new Date() ));
                out.write( " - " );
                out.write( desc );
                out.write( lineSeparator );
                out.write( s );
                out.write( lineSeparator );
                out.flush();
            }
        } catch( InterruptedIOException x ) {
            Thread.currentThread().interrupt();
        } catch( IOException ioe ) {
            trouble = true;
        }
    }

/** 
 * Print the string representation of an <code>Object</code> to the screen.
 *
 * @param desc a general description of the source of the message
 * @param o the <code>Object</code> to print
 */
 
    public void println( String desc, Object o ) {
        try {
            synchronized( lock ) {
                ensureOpen();
                out.write( sdf.format( new Date() ));
                out.write( " - " );
                out.write( desc );
                out.write( lineSeparator );
                out.write( String.valueOf( o ));
                out.write( lineSeparator );
                out.flush();
            }
        } catch( InterruptedIOException x ) {
            Thread.currentThread().interrupt();
        } catch( IOException ioe ) {
            trouble = true;
        }
    }
}

