/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <mballen@erols.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package jcifs.util;

import java.util.Properties;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Date;

/**
 *
 * 
 * This class is backwords compatible with {@link java.util.Properties}
 * however another dimension is added by using the customary dot '.' as
 * an operator to alternativly represent properties as trees. In essance
 * it gives meaning to the '.' used in keys of properties whereas with
 * traditional properties usage of the dot '.' was mearly a convention. This
 * allows applications to organize there data in new ways. The following
 * is an example of the "tagged" output from a PropertiesTree along with
 * traditional output from the same PropertiesTree to it's right. PropertiesTree can load and store
 * both formats at runtime(e.g. load traditional and then store tagged).
 * 
 * <p><table border="1" cellpadding="3" cellspacing="0" width="100%"><tr><td>
 * <pre>
 * #Feed Processor Config
 * #Sun Dec 31 00:48:23 EST 2000
 *
 * proxy=192.168.1.15
 * &lt;net&gt;
 *     username=joe
 *     &lt;smb&gt;
 *         host=doc-storage
 *         username=kelly
 *     &lt;/smb&gt;
 *     &lt;ftp&gt;
 *         host=feed10.research.com
 *     &lt;/ftp&gt;
 * &lt;/net&gt;
 * </pre>
 * 
 * </td><td>
 * 
 * <pre>
 * #Feed Processor Config
 * #Sun Dec 31 00:59:31 EST 2000
 *
 * proxy=192.168.1.15
 * net.username=joe
 * net.smb.host=doc-storage
 * net.smb.username=kelly
 * net.ftp.host=feed10.research.com
 * </pre>
 * 
 * </td></tr></table>
 *
 * <p>In the above example, if <code>getProperty( "net.ftp.username" )</code>
 * was called "joe" would be returned. The <code>get</code> method walks
 * up the tree searching for the first match. So the following:
 * 
 * <pre>
 *     net.ftp.username
 *     net.username
 *     username
 * </pre>
 * 
 * are basically equivalent and thus backwords compatible with
 * java.util.Properties. The plus is that a child property can mask a parent
 * property. This is true of the <code>net.smb.username</code> property.
 * It might also be thought of as overloading <code>net.username</code>
 * with <code>net.smb.username</code> but only for the smb branch.
 *
 */

public class PropertiesTree extends Properties {

	private static final String INDENT = "    ";
	private static final int TRUNC = 70;

	private String label;
	private PropertiesTree parent;

	private void listTagged( String indent, PrintStream out ) {
		Vector sections = new Vector(0);
		for( Enumeration e = keys(); e.hasMoreElements(); ) {
			String key = (String)e.nextElement();
			Object value = get( key );
			if( value instanceof PropertiesTree ) {
				sections.addElement( value );
			} else {
				String s;
				s = indent + key + "=" + (String)value;
				s = s.length() > TRUNC ? s.substring( 0, TRUNC - 3 ) + "..." : s;
				out.println( s );
			}
		}
		for( Enumeration e1 = sections.elements(); e1.hasMoreElements(); ) {
			String s;
			PropertiesTree p = (PropertiesTree)e1.nextElement();

			s = indent + "<" + p.label + ">";
			s = s.length() > TRUNC ? s.substring( 0, TRUNC - 3 ) + "..." : s;
			out.println( s );

			p.list0( indent + INDENT, out );

			s = indent + "</" + p.label + ">";
			s = s.length() > TRUNC ? s.substring( 0, TRUNC - 3 ) + "..." : s;
			out.println( s );
		}
	}
	private void list0( String prefix, PrintStream out ) {
		Vector sections = new Vector(0);
		for( Enumeration e = keys(); e.hasMoreElements(); ) {
			String key = (String)e.nextElement();
			Object value = get( key );
			if( value == null ) {
			} else if( value instanceof PropertiesTree ) {
				sections.addElement( value );
			} else {
				if( prefix != null ) {
					key = prefix + "." + key;
				}
				String s = (String)value;
				s = s.length() > TRUNC ? s.substring( 0, TRUNC - 3 ) + "..." : s;
				out.println( key + "=" + s );
			}
		}
		for( Enumeration e1 = sections.elements(); e1.hasMoreElements(); ) {
			PropertiesTree p = (PropertiesTree)e1.nextElement();
			p.list0( prefix == null ? p.label : prefix + "." + p.label, out );
		}
	}
	private synchronized void saveTagged( String indent, PrintWriter out ) {
		Vector sections = new Vector(0);
		for( Enumeration e = keys(); e.hasMoreElements(); ) {
			String key = (String)e.nextElement();
			Object value = get( key );
			if( value instanceof PropertiesTree ) {
				sections.addElement( value );
			} else if( value instanceof String ) {
				out.print( indent );
				out.print( key );
				out.write( '=' );

				String val = (String)value;
				int len = val.length();
				boolean empty = false;

				for (int i = 0 ; i < len ; i++) {
					int ch = val.charAt(i);

					switch (ch) {
						case '\\': out.write('\\'); out.write('\\'); break;
						case '\t': out.write('\\'); out.write('t'); break;
						case '\n': out.write('\\'); out.write('n'); break;
						case '\r': out.write('\\'); out.write('r'); break;

						default:
							if ((ch < ' ') || (ch >= 127) || (empty && (ch == ' '))) {
								if( ch > 255 ) {
									out.write(ch);
								} else {
									out.write('\\');
									out.write('u');
									out.write(toHex((ch >> 12) & 0xF));
									out.write(toHex((ch >>  8) & 0xF));
									out.write(toHex((ch >>  4) & 0xF));
									out.write(toHex((ch >>  0) & 0xF));
								}
							} else {
								out.write(ch);
							}
					}
					empty = false;
				}
				out.write('\n');
				out.flush();
			} else {
				System.out.println( "error: " + key + "=" + value );
			}
		}
		for( Enumeration e1 = sections.elements(); e1.hasMoreElements(); ) {
			PropertiesTree p = (PropertiesTree)e1.nextElement();
			out.println( indent + "<" + p.label + ">" );
			p.saveTagged( indent + INDENT, out );
			out.println( indent + "</" + p.label + ">" );
		}
	}
	private synchronized void save0( String prefix, PrintWriter out ) {
		Vector sections = new Vector(0);
		for( Enumeration e = keys(); e.hasMoreElements(); ) {
			String key = (String)e.nextElement();
			Object value = get( key );
			if( value instanceof PropertiesTree ) {
				sections.addElement( value );
			} else if( value instanceof String ) {
				if( prefix != null ) {
					out.print( prefix );
					out.write( '.' );
				}
				out.print( key );
				out.write( '=' );

				String val = (String)value;
				int len = val.length();
				boolean empty = false;

				for (int i = 0 ; i < len ; i++) {
					int ch = val.charAt(i);

					switch (ch) {
						case '\\': out.write('\\'); out.write('\\'); break;
						case '\t': out.write('\\'); out.write('t'); break;
						case '\n': out.write('\\'); out.write('n'); break;
						case '\r': out.write('\\'); out.write('r'); break;

						default:
							if ((ch < ' ') || (ch >= 127) || (empty && (ch == ' '))) {
								if( ch > 255 ) {
									out.write(ch);
								} else {
									out.write('\\');
									out.write('u');
									out.write(toHex((ch >> 12) & 0xF));
									out.write(toHex((ch >>  8) & 0xF));
									out.write(toHex((ch >>  4) & 0xF));
									out.write(toHex((ch >>  0) & 0xF));
								}
							} else {
								out.write(ch);
							}
					}
					empty = false;
				}
				out.write('\n');
				out.flush();
			} else {
				System.out.println( "error: " + key + "=" + value );
			}
		}
		for( Enumeration e1 = sections.elements(); e1.hasMoreElements(); ) {
			PropertiesTree p = (PropertiesTree)e1.nextElement();
			p.save0( prefix == null ? p.label : prefix + "." + p.label, out );
		}
	}
	private static char toHex(int nibble) {
		return hexDigit[(nibble & 0xF)];
	}
	private static char[] hexDigit = {
		'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'
	};

	/**
	 * Construct an empty properties tree with no values.
	 */

	public PropertiesTree() {
		this.label = null;
		this.parent = null;
	}

	/** 
	 * Put a key and value into the tree. If the key contains a dot
	 * '.', new tree nodes will be automatically created to accomodate.
	 */ 

	public synchronized Object put( Object key, Object value ) {
		String s = (String)key;
		String l;
		Object o;
		PropertiesTree p;
		int d;
		if(( d = s.indexOf( '.' )) > -1 ) {
			l = s.substring( 0, d );
			o = super.get( l );
			if( o != null && o instanceof PropertiesTree ) {
				p = (PropertiesTree)o;
			} else if( o != null ) {
				throw new IllegalArgumentException( "property exists: " +
											l + "=" + (String)o );
			} else {
				p = new PropertiesTree();
				p.label = l;
				p.parent = this;
				super.put( l, p );
			}
			return p.put( s.substring( d + 1 ), value );
		} else {
			o = super.get( s );
			if( o != null && o instanceof PropertiesTree ) {
				throw new IllegalArgumentException( "property tree exists: " + s );
			}
			if( value instanceof PropertiesTree ) {
				((PropertiesTree)value).label = (String)key;
				((PropertiesTree)value).parent = this;
			}
			return super.put( key, value );
		}
	}

	/** 
	 * Retrieved the property specified by the key parameter, seraching parent
	 * nodes if necessary.
	 */ 

	public synchronized Object get( Object key ) {
		String s = (String)key;
		int d;
		if(( d = s.indexOf( '.' )) > 0 ) {
			Object o = super.get( s.substring( 0, d ));
			if( o == null ) {
				return get( s.substring( d + 1 ));
			} else if( o instanceof PropertiesTree ) {
				return ((PropertiesTree)o).get( s.substring( d + 1 ));
			}
		} else {
			Object o = super.get( s );
			if( o == null && parent != null ) {
				return parent.get( s );
			} else if( o instanceof PropertiesTree ||
					(  o instanceof String && ((String)o).length() > 0 )) {
				return o;
			}
		}
		return null;
	}

	/**
	 * Retrieve a property from the tree by searching parent nodes
	 * if necessary and return the value as a String.
	 */

	public String getProperty( String key ) {
		Object o = get( key );
		return o == null ? null : (String)o;
	}

	/** 
	 * Retrieve a property from the tree by searching parent nodes if necessary
	 * and return the value as a String. If the keys value is not found, the
	 * default parameter will be returned.
	 */ 

	public String getProperty( String key, String defaultValue ) {
		Object o = get( key );
		return o == null ? defaultValue : (String)o;
	}

	/** 
	 * List all properties in the traditional output format and truncated to
	 * ensure the contents fit within the display.
	 */ 

	public void list( PrintStream out ) {
		out.println("-- listing properties tree --");
		list0( null, out );
	}

	/** 
	 * Save this properties tree to the specified OutputStream. The header
	 * parameter will be printed at the top as a comment.
	 */ 

	public synchronized void save( OutputStream out, String header ) {
		save( out, header, false );
	}

	/** 
	 * Save this properties tree to the specified OutputStream. The header
	 * parameter will be printed at the top as a comment. If the boolean
	 * parameter is true the output will be stored in tagged format as shown
	 * in the example at the top of this page.
	 */ 

	public synchronized void save( OutputStream out, String header, boolean tagged ) {
		PrintWriter prnt = new PrintWriter( out, true );
		if( header != null ) {
			prnt.write( '#' );
			prnt.println( header );
		}
		prnt.write( '#' );
		prnt.println( new Date() );
		if( tagged ) {
			saveTagged( "", prnt );
		} else {
			save0( null, prnt );
		}
	}

	/** 
	 * Load all properties from the provided InputStream. The text read may be
	 * either tagged format, traditional format, or a mixture of both.
	 */ 

	public synchronized void load( InputStream in ) throws IOException {
		int buflen = 80;
		char[] buf = new char[buflen];
		int bufindx = 0;
		boolean intag = false;
		String section = null;

		int ch = in.read();
		while( true ) {
			switch( ch ) {
				case -1:
					if( section != null ) {
						throw new IOException( "expecting end tag" );
					}
					return;
				case '<':
					intag = true;
					ch = in.read();
					continue;
				case '#':
				case '!':
					do {
						ch = in.read();
					} while(( ch >= 0 ) && ( ch != '\n' ) && ( ch != '\r' ));
					continue;

				case '\n':
				case '\r':
				case ' ':
				case '\t':
					ch = in.read();
					continue;
			}

			bufindx = 0;
			while(( ch >= 0 ) && ( ch != '=' ) && ( ch != ':' ) && ( ch != ' ' ) &&
								( ch != '\t' ) && ( ch != '\n' ) && ( ch != '\r' )) {
				if( intag && ch == '>' ) {
					ch = in.read();
					break;
				}
				if( bufindx >= buflen ) {
					buflen *= 2;
					char[] nbuf = new char[buflen];
					System.arraycopy( buf, 0, nbuf, 0, buf.length );
					buf = nbuf;
				}
				buf[bufindx++] = (char)ch;
				ch = in.read();
			}
			String key = new String( buf, 0, bufindx );
			if( intag ) {
				if( key.startsWith( "/" )) {
					int s;
					// chop off end tag from current section
					if( section != null && section.regionMatches(
							( s = section.lastIndexOf( '.' )) + 1, key, 1, key.length() - 1 )) {
						section = s < 1 ? null : section.substring( 0, s );
					} else {
						throw new IOException( "unexpected end tag: <" + key + ">" );
					}
				} else {
					section = section == null ? key : section + "." + key;
				}
				intag = false;
				continue;
			}

			while(( ch == ' ' ) || ( ch == '\t' )) {
				ch = in.read();
			}
			if(( ch == '=' ) || ( ch == ':' )) {
				ch = in.read();
			}
			while(( ch == ' ' ) || ( ch == '\t' )) {
				ch = in.read();
			}

			bufindx = 0;
			while(( ch >= 0 ) && ( ch != '\n' ) && ( ch != '\r' )) {
				int next = 0;
				if( ch == '\\' ) {
					switch( ch = in.read() ) {
						case '\r':
							if((( ch = in.read() ) == '\n' ) ||
									( ch == ' ' ) || ( ch == '\t' )) {
							} else continue;
						case '\n': 
							while((( ch = in.read() ) == ' ') || ( ch == '\t' ));
							continue;
						case 't': ch = '\t'; next = in.read(); break;
						case 'n': ch = '\n'; next = in.read(); break;
						case 'r': ch = '\r'; next = in.read(); break;
						case 'u': {
							while ((ch = in.read()) == 'u');
							int d = 0;
							loop:
								for (int i = 0 ; i < 4 ; i++) {
									next = in.read();
									switch (ch) {
										case '0': case '1': case '2': case '3': case '4':
										case '5': case '6': case '7': case '8': case '9':
											d = (d << 4) + ch - '0';
											break;
										case 'a': case 'b': case 'c':
										case 'd': case 'e': case 'f':
											d = (d << 4) + 10 + ch - 'a';
											break;
										case 'A': case 'B': case 'C':
										case 'D': case 'E': case 'F':
											d = (d << 4) + 10 + ch - 'A';
											break;
										default:
											break loop;
									}	
									ch = next;
								}
								ch = d;
								break;
						}
						default:
							next = in.read();
							break;
					}
				} else {
					next = in.read();
				}
				if (bufindx >= buflen) {
					buflen *= 2;
					char[] nbuf = new char[buflen];
					System.arraycopy(buf, 0, nbuf, 0, buf.length);
					buf = nbuf;
				}
				buf[bufindx++] = (char)ch;
				ch = next;
			}
			String val = new String(buf, 0, bufindx);

			if( section != null ) {
				put( section + '.' + key, val );
			} else {
				put( key, val );
			}
		}
	}
}
